//root)/Gen7/HandheldSourceCode/Overlay/recipes/g7common/g7client.c - Rev 11868
// Rev 11865 | Blame | Compare with Previous | Last modification | View Log | RSS feed
// START
// Gen7 Audio Client Library
// Copyright (c) 2010-2015 Mimosa Acoustics, Inc.

// Linux end-of-line markers, 8-column _tabs_ (not spaces), K&R C brace style.

// SOURCE CONTROL WARNING:  The master version of this file is located in the Gen7 repository at
//      Gen7/HandheldSourceCode/Overlay/recipes/g7common/g7client.c
// Subversion updates under Windows may result in duplicate copies due to following symbolic links,
// but all changes should be made and committed to the master version at the location showmn above.

// For Dev Studio 2005, Had to add an Additional Dependency to the linker
// go to Project-> <Project Name> Properties -> Configuration Properties -> Linker ->
// Input and add Ws2_32.lib to the Additional Dependencies.

#include <stdio.h>
#include <string.h>

// Define MAXPATHLEN based on platform-specific equivalents.
#if defined(_MSC_VER)
#define MAXPATHLEN      _MAX_PATH
#define _CRT_SECURE_NO_WARNINGS 1
#pragma warning(disable: 4996)
#else
#include <sys/param.h>          // for PATH_MAX
#define MAXPATHLEN      PATH_MAX
// Windows wants POSIX-style _strdup; Linux hasn't caught up.
#define _strdup strdup
#endif

// These should be defined externally (e.g., in the Makefile) if you don't like the defaults.
// A combination of G7CLIENT_DEBUG_x bits which are compiled in and can be be enabled at run-time.
#ifndef G7CLIENT_COMPILE_DEBUG
#define G7CLIENT_COMPILE_DEBUG  0xFFFF
#endif                                

// For MINGW compilation or Visual C++ compilation, use the Windows Sockets version of network calls.
#if defined(MINGW) || defined(_MSC_VER)
#define USE_WINSOCK
#else
#undef USE_WINSOCK
#endif

#if defined(USE_WINSOCK)
// Windows
#include <winsock2.h>
#include <stdarg.h>
typedef int socklen_t;
#else
// POSIX
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdarg.h>
#include <malloc.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

typedef int SOCKET;
#define SOCKET_ERROR -1
#define INVALID_SOCKET -1
#define closesocket close
#endif

//  Windows can't deal with symbolic links, so we have to know where to find these files.
#if defined(_MSC_VER)
#include "../g7common/g7shared.h"
#include "../g7common/g7client.h"
#else
#include "g7shared.h"
#include "g7client.h"
#endif

// Defaults, all reset during g7_init.
#define DEFAULT_CONNECT_MS              2000
#define DEFAULT_PROGRESS_MS             2000
#define DEFAULT_SAMPLES_PER_BLOCK       1024    // samples per block

// The number of active instances.
// If 0, Windows Sockets (if applicable) will be initialized by the first created instance.
// If 1, Windows Sockets (if applicable) will be shut down along with the next shut-down instance.
static int active_instances = 0;

// Characters in each error string buffer.
#define ERRBUF_SIZE     1024

struct instance {
        // A description of the last top-level error which occurred.
        char main_errortext[ERRBUF_SIZE];

        // A buffer used to generate system error messages.
        char work_errortext[ERRBUF_SIZE];

        // Which messages we are logging (none by default).
        // This is a combination of the G7CLIENT_DEBUG_ bit constants.
        unsigned int client_debug_mask;

        // The path name of the log file.
        char log_path[MAXPATHLEN + 1];

        // The number of milliseconds we wait for the peripheral to accept a connection.
        // NOTE: setting this to 0 disables timeouts (e.g., for debugging).
        unsigned int connect_ms;

        // The number of milliseconds we wait for the peripheral to provide the next installment
        // of data (when reading)or to accept the next installment of data (when writing).
        // NOTE: setting this to 0 disables timeouts (e.g., for debugging).
        unsigned int progress_ms;

        // The block size in samples.
        unsigned int samples_per_block;

        // The client socket connected to the server.
        SOCKET client;
};

static void debug_internal(struct instance* h_, const char* format_, va_list args_)
{
        char text[1024];
        int saved_errno;
        FILE* out;

        // NUL-terminate in case vsnprintf fails.
        text[0] = '\0';

        // Preserve errno across the library calls we are about to make.
        saved_errno = errno;
#if defined(_MSC_VER)
        vsnprintf_s(text, 1024, _TRUNCATE, format_, args_);
#else
        vsnprintf(text, 1024, format_, args_);
#endif

        // Where are we writing?
        out = stdout;
        if (h_->log_path[0] != '\0') {
                // If this fails, we can't do much about it; presumably they don't want output to go to stdout
                // or they wouldn't have specified a log path in the first place, so we don't want to pound out
                // a message to stdout on every logging attempt.  So we just fail silently and discard the output.
#if defined(_MSC_VER)
                fopen_s(&out, h_->log_path, "a");
#else
                out = fopen(h_->log_path, "a");
#endif
        }
        if (out != NULL) {
                fputs(text, out);
                fputs("\r\n", out);
        }
        if ((out != NULL) && (out != stdout)) {
                // Close the file (except stdout) after every write so it is always flushed and always accessible.
                fclose(out);
        }

        // Restore the original errno value.
        errno = saved_errno;
}

static void debug(struct instance* h_, const char* format_, ...)
{
        va_list(args);
        va_start(args, format_);
        debug_internal(h_, format_, args);
}

#if (G7CLIENT_COMPILE_DEBUG & G7CLIENT_DEBUG_NETWORK)
#define DBG_NETWORK(H_, ...)
        if (H_->client_debug_mask & G7CLIENT_DEBUG_NETWORK) { debug(H_, __VA_ARGS__); }
#else
#define DBG_NETWORK(...)
#endif

#if (G7CLIENT_COMPILE_DEBUG & G7CLIENT_DEBUG_NETWORK)
#define DBG_CALLS(H_, ...)
        if (H_->client_debug_mask & G7CLIENT_DEBUG_CALLS) { debug(H_, __VA_ARGS__); }
#else
#define DBG_CALLS(...)
#endif
// Internal methods.

// Get a system error code.
static int last_error()
{
#if defined(USE_WINSOCK)
        return WSAGetLastError();
#else
        return errno;
#endif
}

// Return a pointer to a description of an error returned by last_error()
// The pointer may be a constant pointer or may point to the work_errortext buffer in the instance.
static const char* errortext(struct instance* h_, int error_)
{
#if defined(USE_WINSOCK)
        switch (error_) {
        case WSANOTINITIALISED:
                return "not initialized with WSAStartup";
        case WSAENETDOWN:
                return "network down";
        case WSAEACCES:
                return "broadcast address used without SO_BROADCAST";
        case WSAEINTR:
                return "canceled with WSACancelBlockingCall";
        case WSAEINPROGRESS:
                return "blocking call in progress";
        case WSAEFAULT:
                return "illegal \"buf\" parameter";
        case WSAENETRESET:
                return "keep-alive failure";
        case WSAENOBUFS:
                return "no buffer space available";
        case WSAENOTCONN:
                return "socket not connected";
        case WSAENOTSOCK:
                return "descriptor not a socket";
        case WSAEOPNOTSUPP:
                return "MSG_OOB improperly specified";
        case WSAESHUTDOWN:
                return "socket shut down";
        case WSAEWOULDBLOCK:
                return "blocking operation on nonblocking socket";
        case WSAEMSGSIZE:
                return "message too large";
        case WSAEHOSTUNREACH:
                return "remote host unreachable";
        case WSAEINVAL:
                return "unbound socket or improper flag";
        case WSAECONNABORTED:
                return "connection aborted";
        case WSAECONNRESET:
                return "connection reset by peer";
        case WSAETIMEDOUT:
                return "timed out";
        case WSAEMFILE:
                return "too many open files";
        case WSAEALREADY:
                return "operation already in progress";
        case WSAEDESTADDRREQ:
                return "address omitted";
        case WSAEPROTOTYPE:
                return "wrong protocol type";
        case WSAENOPROTOOPT:
                return "improper protocol option";
        case WSAEPROTONOSUPPORT:
                return "unsupported protocol";
        case WSAESOCKTNOSUPPORT:
                return "unsupported socket type";
        case WSAEPFNOSUPPORT:
                return "unsupported protocol family";
        case WSAEAFNOSUPPORT:
                return "unsupported address family";
        case WSAEADDRINUSE:
                return "address already in use";
        case WSAEADDRNOTAVAIL:
                return "unassignable address";
        case WSAENETUNREACH:
                return "network unreachable";
        case WSAEISCONN:
                return "socket already connected";
        case WSAECONNREFUSED:
                return "connection refused";
        case WSAEPROCLIM:
                return "too many processes";
        case WSAEHOSTDOWN:
                return "remote host down";
        case WSASYSNOTREADY:
                return "network subsystem unavailable";
        case WSAVERNOTSUPPORTED:
                return "improper winsock.dll version";
        case WSAEDISCON:
                return "shutdown in progress";
        case WSAHOST_NOT_FOUND:
                return "host not found";
        case WSATYPE_NOT_FOUND:
                return "class type not found";
        case WSATRY_AGAIN:
                return "nonauthoritative host not found";
        case WSANO_RECOVERY:
                return "nonrecoverable error";
        case WSANO_DATA:
                return "no data record";

        // OS-dependent.
        case WSA_INVALID_HANDLE:
                return "invalid event handle";
        case WSA_INVALID_PARAMETER:
                return "invalid parameter";
        case WSA_IO_INCOMPLETE:
                return "IO incomplete";
        case WSA_IO_PENDING:
                return "IO pending";
        case WSA_NOT_ENOUGH_MEMORY:
                return "insufficient memory";
        case WSA_OPERATION_ABORTED:
                return "overlapped operation aborted";
//      case WSAINVALIDPROCTABLE:
//              return "invalid procedure table";
//      case WSAINVALIDPROVIDER:
//              return "invalid service provider version";
//      case WSAPROVIDERFAILEDINIT:
//              return "service provider unitializable";
        case WSASYSCALLFAILURE:
                return "system call failure";
        }

        // Not matched.
#if defined(_MSC_VER)
        _snprintf_s(h_->work_errortext, ERRBUF_SIZE, _TRUNCATE, "unknown error: 0x%08X", error_);
#else
        _snprintf(h_->work_errortext, ERRBUF_SIZE, "unknown error: 0x%08X", error_);
#endif
        return h_->work_errortext;
#else
        strerror_r(error_, h_->work_errortext, ERRBUF_SIZE);
        return h_->work_errortext;
#endif
}

// Return text describing the last network error.
static const char* last_errortext(struct instance* h_)
{
        return errortext(h_, last_error());
}

// Return text describing the error reported by a given socket via getsockopt SO_ERROR.
static const char* socket_errortext(struct instance* h_, SOCKET socket_)
{
        int ret;
        int error;
        int len;

        // Get and clear the last-error indication from the specified socket.
        len = sizeof(error);
        ret = getsockopt(socket_, SOL_SOCKET, SO_ERROR, (char*)&error, (socklen_t*)&len);
        if (ret == SOCKET_ERROR) {
                // Failed to get the error code itself.
                return "Unknown error (SO_ERROR unavailable)";
        }

        // Interpret the error reported by the socket.
        return errortext(h_, error);
}

static void set_errtext_internal(struct instance* h_, const char* format_, va_list args_)
{
        // NULL pointer just clears the error.
        if (format_ == NULL) {
                strcpy(h_->main_errortext, "No error");
                return;
        }

// Call the real format method.
#if defined(USE_WINSOCK)
#if defined(_MSC_VER)
        if (0 >= _vsnprintf_s(h_->main_errortext, 1024, _TRUNCATE, format_, args_)) {
#else
        if (0 >= _vsnprintf(h_->main_errortext, 1024, format_, args_)) {
#endif
#else
        if (0 >= vsnprintf(h_->main_errortext, 1024, format_, args_)) {
#endif
                strcpy(h_->main_errortext, "Error description not available");
        }
}

// Wrapper for setting error text with printf semantics.
void set_errtext(struct instance* h_, const char* format_, ...)
{
        va_list args;
        va_start(args, format_);
        set_errtext_internal(h_, format_, args);
}

// External function for setting error text with printf semantics (for helper functions).
void g7_set_errtext(g7handle_t handle_, const char* format_, ...)
{
        va_list args;

        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;
        if (h == NULL) {
                return;
        }

        va_start(args, format_);
        set_errtext_internal(h, format_, args);
}

// Called when a function is about to return a result.
// If format_ is non-NULL, errtext is set to the specified format string and printf-style arguments.
// If format_ is non-NULL and DBG_CALL logging is enabled, the message is logged.
// Returns the specified error code, so this can be used in a return statement.
static int pre_return(struct instance* h_, const char* function_, int error_, const char* format_, ...)
{
        // Format and set errtext, unless the caller wants to log the error text itself.
        if ((format_ != NULL) && (format_ != h_->main_errortext)) {
                va_list args;
                va_start(args, format_);
                set_errtext_internal(h_, format_, args);
        }

        // Log a debug message.
        if (format_ != NULL) {
                DBG_CALLS(h_, "%s: %s: %s", function_, g7_returntext(h_, error_), h_->main_errortext);
        } else {
                DBG_CALLS(h_, "%s: %s", function_, g7_returntext(h_, error_));
        }

        // Pass through the error code from the caller.
        return error_;
}

// Send (receive) the specified data out on a non-blocking socket.
// send_ is 1 to send, 0 to receive.
// socket_ is the socket descriptor.
// data_ is the buffer of data to be sent (received).
// bytes_ is the number of bytes to send (receive).  There are no partial reads (writes);
//      a successful return means that all bytes_ bytes were sent (received) before timing out.
// Returns G7_OK or a G7_ERR_ code (with errtext set appropriately) on error.
// The socket NOT closed if an error occurs; the caller must do that.
static int traffic(struct instance* h_, const char* caller_, int send_, SOCKET socket_, char* data_, int bytes_)
{
        struct timeval tv;
        struct timeval* use_timeout;

        // What are we doing?
        const char* act = send_ ? "sending" : "receiving";

        // Determine the timeout period as a timeval structure.
        use_timeout = NULL;
        if (h_->progress_ms != 0) {
                tv.tv_sec = h_->progress_ms / 1000;
                tv.tv_usec = (h_->progress_ms - (tv.tv_sec * 1000)) * 1000;
                use_timeout = &tv;
        }

        DBG_NETWORK(h_, "Entering traffic loop for %s %d bytes", act, bytes_);
        for (;;) {
                // Set up to select, waiting for writability (readability) or an exception condition.
                fd_set rwfds;
                fd_set exceptfds;
                int ret;
                FD_ZERO(&rwfds);
                FD_ZERO(&exceptfds);
#if defined(USE_WINSOCK)
                rwfds.fd_count = 1;
                rwfds.fd_array[0] = socket_;
                exceptfds.fd_count = 1;
                exceptfds.fd_array[0] = socket_;
#else
                FD_SET(socket_, &rwfds);
                FD_SET(socket_, &exceptfds);
#endif

                // Wait.
                // The first parameter to select() (highest socket descriptor in any set) is for POSIX, not Windows.
                if (send_) {
                        // waiting for writability or exception.
                        ret = select((int)socket_ + 1, NULL, &rwfds, &exceptfds, use_timeout);
                } else {
                        // waiting for readability or exception.
                        ret = select((int)socket_ + 1, &rwfds, NULL, &exceptfds, use_timeout);
                }
                if (ret == SOCKET_ERROR) {
                        DBG_NETWORK(h_, "  select error: %s", last_errortext(h_));
                        closesocket(socket_);
                        return pre_return(h_, caller_, G7_ERR_NETWORK, "Error waiting for audio device while %s data: %s", act, last_errortext(h_));
                }
                if (ret == 0) {
                        // The timeout expired.
                        DBG_NETWORK(h_, "  timed out");
                        return pre_return(h_, caller_, G7_ERR_NETWORK, "Timed out %s network traffic", act);
                }
                if (FD_ISSET(socket_, &exceptfds)) {
                        // The connection dropped.
                        DBG_NETWORK(h_, "  exceptional condition on socket");
                        return pre_return(h_, caller_, G7_ERR_NETWORK, "Exceptional condition on socket");
                }

                // Send the next chunk of data.  ret will be the number of bytes actually sent.
                if (send_) {
#if defined(_MSC_VER)
                        ret = send(socket_, data_, bytes_, 0);
#else
                        ret = send(socket_, data_, bytes_, MSG_NOSIGNAL);
#endif
                } else {
                        ret = recv(socket_, data_, bytes_, 0);
                }
                if (ret > 0) {
                        DBG_NETWORK(h_, "  finished %s %d bytes", act, ret);
                }

                if (ret == bytes_) {
                        // We sent (received) the entire remainder of the request.
                        break;
                }
                // Use < 0 rather than SOCKET_ERROR to avoid disaster on an unexpected negative return.
                if (ret < 0) {
                        // Check for a real error (anything other than "would block").
                        int error = last_error();
#if defined(USE_WINSOCK)
                        if (error != WSAEWOULDBLOCK) {
#else
                        if ((error != EWOULDBLOCK) && (error != EAGAIN)) {
#endif
                                return pre_return(h_, caller_, G7_ERR_NETWORK, "Error %s network traffic: %s", act, last_errortext(h_));
                        }
                        // WSAEWOULDBLOCK falls through and continues waiting.
                } else if (ret == 0) {
                        // Since we selected, we shouldn't get here and have the socket unable to read or write;
                        // that indicates that the connection was closed by the remote end.
                        DBG_NETWORK(h_, "  socket was closed");

                        // The connection dropped.
                        return pre_return(h_, caller_, G7_ERR_NETWORK, "Connection to audio device was closed");
                } else if (ret < bytes_) {
                        // A partial write (read) is normal, and just means we have more to send (receive).
                        // Advance our position in the data buffer, decrease the remaining bytes, and keep waiting.
                        data_ += ret;
                        bytes_ -= ret;
                }
        }

        // Success.  Note that there is no pre_return call here--the caller will do any logging.
        return G7_OK;
}

// Send a request to the audio device and receive the response.
// request_ is a pointer to the request header.  The code, argX, and following_bytes fields
//      must be set (in host byte order).  following_bytes must be 0 if request_data_ = NULL.
//
// request_data_ is a pointer to any data to be sent following the request header
//      (the number of bytes is specified by the following_bytes field in the request header).
//
// response_ is a pointer to where a successful response header should be placed.  It may be
//      NULL if the caller doesn't want to receive the response header.
//
// response_data_ is a pointer to where successful response data should be placed.  It may be NULL,
//      in which case any data received from the device is discarded.
//
// response_data_bytes_ points to an in/out parameter which must be set to the the number of
//      bytes available at response_data_ prior to the call; on return it will be set to the
//      number of bytes actually stored at response_data_.  If the audio device returns more
//      than response_data_bytes_ bytes, the remainder of the response will be discarded.
//      It may be NULL, which is the same as providing a pointer to a zero--any data received
//      from the device will be discarded.
//
// If the call fails (including local failures or failures reported by the audio device),
// errtext will be set appropriately and a G_ERR_ code will be returned.  In this case, the
// buffers pointed to by response_ and response_data_ will not be affected; i.e., the caller
// will receive neither header information nor data from a failing request.
//
// Returns G_OK on success; G_ERR_ otherwise.
static int issue(struct instance* h_, const char* caller_, const struct msg_t* request_, const char* request_data_,
        struct msg_t* response_, char* response_data_, unsigned int* response_data_bytes_)
{
        int ret;                        // general return from system calls
        unsigned int rx_bytes;          // computed bytes of response data to receive
        struct msg_t local_request;     // local copy of request in network byte order
        struct msg_t local_response;    // local copy of response in host byte order

        // Must have a connection.
        if (h_->client == INVALID_SOCKET) {
                return pre_return(h_, caller_, G7_ERR_STATE, "No connection to server");
        }

        // Convert and send the header.
        local_request.magic = htonl(G7_MAGIC);
        local_request.code = htonl(request_->code);
        local_request.arg1 = htonl(request_->arg1);
        local_request.arg2 = htonl(request_->arg2);
        local_request.arg3 = htonl(request_->arg3);
        local_request.arg4 = htonl(request_->arg4);
        local_request.following_bytes = htonl(request_->following_bytes);
        // We can't use any fields in local_request after this because they may be in the wrong byte order.
        ret = traffic(h_, caller_, 1, h_->client, (char*)&local_request, sizeof(local_request));
        if (ret < 0) {
                closesocket(h_->client);
                h_->client = INVALID_SOCKET;

                // Do not call pre_return; traffic() will already have done that.
                return ret;
        }

        // Write any following data.
        if (request_->following_bytes > 0) {
                ret = traffic(h_, caller_, 1, h_->client, (char*)request_data_, request_->following_bytes);
                if (ret != G7_OK) {
                        closesocket(h_->client);
                        h_->client = INVALID_SOCKET;

                        // Do not call pre_return; traffic() will already have done that.
                        return ret;
                }
        }

        // Receive the reply header into a local buffer.
        ret = traffic(h_, caller_, 0, h_->client, (char*)&local_response, sizeof(local_response));
        if (ret != G7_OK) {
                closesocket(h_->client);
                h_->client = INVALID_SOCKET;

                // Do not call pre-return; traffic() will already have done that.
                return ret;
        }

        // Fix the byte ordering in the response.
        local_response.magic = ntohl(local_response.magic);
        local_response.code = ntohl(local_response.code);
        local_response.arg1 = ntohl(local_response.arg1);
        local_response.arg2 = ntohl(local_response.arg2);
        local_response.arg3 = ntohl(local_response.arg3);
        local_response.arg4 = ntohl(local_response.arg4);
        local_response.following_bytes = ntohl(local_response.following_bytes);
        // We can now use values from local_response since they're in host byte order.

        // Did we get junk instead of a legitimate response?
        if (local_response.magic != G7_MAGIC) {
                closesocket(h_->client);
                h_->client = INVALID_SOCKET;
                return pre_return(h_, caller_, G7_ERR_NETWORK, "Received corrupted response from audio device");
        }

        // Did the client report back with just a header, and possibly a message?
        if ((local_response.code == G7_MSG_ACK) || (local_response.code == G7_MSG_NAK)) {

                // Get a pointer to any text returned by the server.
                char* text_buffer = NULL;
                const char* server_text = "Details not available";

                // If there is any data, put it in errtext.  (After a successful response to INITIALIZE, we will get
                // back a banner string in the error text, which isn't really an "error" but is handled the same way).
                if (local_response.following_bytes != 0) {
                        text_buffer = (char*)malloc(local_response.following_bytes + 1);
                        text_buffer[0] = '\0';
                        ret = traffic(h_, caller_, 0, h_->client, text_buffer, local_response.following_bytes);
                        if (ret != G7_OK) {
                                server_text = "Error obtaining details";
                        } else {
                                // Force NUL termination just in case.
                                text_buffer[local_response.following_bytes] = '\0';
                                server_text = text_buffer;
                        }
                }

                // If the client reported an error, use the server response as the error text value and fail.
                if (local_response.code == G7_MSG_NAK) {
                        return pre_return(h_, caller_, local_response.arg1, server_text);
                }

                // Otherwise, set errtext to any returned result.  This is for INITIALIZE, which returns
                // an explanatory banner which the caller can retrieve using g7_errtext after a G7_OK return.
                set_errtext(h_, server_text);
        }

        // Success; if the caller wants a copy of the header, return it to them.
        if (response_ != NULL) {
                memcpy(response_, &local_response, sizeof(local_response));
        }

        // If the client is responding with data and the caller wants it, receive it.
        rx_bytes = 0;
        if ((local_response.code == G7_MSG_DATA) && (response_data_ != NULL) && (response_data_bytes_ > 0)) {

                // Determine how much of the receive data we want.  This is the minimum of the size of the caller's
                // buffer, and the number of bytes of data indicated by the following_bytes field in the header.
                rx_bytes = *response_data_bytes_;
                if (rx_bytes > local_response.following_bytes) {
                        rx_bytes = local_response.following_bytes;
                }

                // Receive any following data.
                if (rx_bytes > 0) {
                        ret = traffic(h_, caller_, 0, h_->client, (char*)response_data_, rx_bytes);
                        if (ret != G7_OK) {
                                closesocket(h_->client);
                                h_->client = INVALID_SOCKET;

                                // Do not call pre-return; traffic() will already have done that.
                                return ret;
                        }
                }
        }

        // Report the number of bytes actually received.
        if (response_data_bytes_ != NULL) {
                *response_data_bytes_ = rx_bytes;
        }

        // Success.
        return G7_OK;
}

// Externally visible methods.

g7handle_t g7_init(unsigned int mask_, const char* log_path_)
{
        // Allocate the instance.
        struct instance* h = (struct instance*)malloc(sizeof(struct instance));
        if (h == NULL) {
                return NULL;
        }
        memset(h, 0, sizeof(struct instance));

        // Set the log path.
        if (log_path_ == NULL) {
                // A NULL log path pointer is the same as an empty log path.
                h->log_path[0] = '\0';
        } else {
#if defined(_MSC_VER)
                strncpy_s(h->log_path, MAXPATHLEN, log_path_, _TRUNCATE);
#else
                strncpy(h->log_path, log_path_, MAXPATHLEN);
#endif
        }

        // Set the debug mask.
        h->client_debug_mask = mask_;

#if defined(USE_WINSOCK)
        // If we are the first instance, initialize Windows Sockets.
        if (active_instances == 0) {
                // Initialize Windows Sockets v2.0.
                WSADATA wsadata;
                int startup = WSAStartup(MAKEWORD(2, 0), &wsadata);
                if (startup != 0) {
                        pre_return(h, "g7_init", G7_ERR_NETWORK, "Could not initialize Windows Sockets (0x%08X)", startup);

                        // Deallocate the instance and return NULL.
                        free(h);
                        return NULL;
                }
        }
#endif

        // Keep track of the number of instances.
        active_instances++;

        // Default timeouts.
        h->connect_ms = DEFAULT_CONNECT_MS;
        h->progress_ms = DEFAULT_PROGRESS_MS;

        // No server connection.
        h->client = INVALID_SOCKET;

        // No block size yet.
        h->samples_per_block = 0;

        // Success; return the instance pointer.
        return (g7handle_t)h;
}

const char* g7_errtext(g7handle_t handle_)
{
        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;
        if (h == NULL) {
                return "NULL instance handle";
        }

        // Return the current contents of the error message.
        return h->main_errortext;
}

const char* g7_returntext(g7handle_t handle_, int code)
{
        // Handle is not currently used.

        // Interpret the code.
        switch (code) {
        case G7_OK:
                return "Success";
        case G7_ERR_INIT:
                return "Not initialized";
        case G7_ERR_STATE:
                return "Audio server state";
        case G7_ERR_NETWORK:
                return "Network error";
        case G7_ERR_ARGS:
                return "Invalid argument";
        case G7_ERR_VERSION:
                return "Version mismatch";
        case G7_ERR_END_OF_STREAM:
                return "End of stream";
        case G7_ERR_SERVER:
                return "Error from server";
        case G7_ERR_NOTCONNECTED:
                return "Not connected";
        case G7_ERR_CONNECTED:
                return "Connected";
        case G7_ERR_SHUTDOWN:
                return "Audio subsystem powered down";

        default:
                return "Unrecognized return value";
        }
}

const char* g7_loopbacktext(g7handle_t handle_, int code_, const char* default_)
{
        // Handle is not currently used.

        // Interpret the loopback code.
        switch (code_) {
        case G7_LOOPBACK_NONE:
                return "none";
        case G7_LOOPBACK_SERVER:
                return "server";
        case G7_LOOPBACK_I2S:
                return "i2s (codec)";
        }
        return default_;
}

int g7_control(g7handle_t handle_, unsigned int flags_, int arg_)
{
        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;

        struct msg_t request;
        int ret = G7_OK;
        unsigned int temporary_progress_ms = 0;
        unsigned int saved_progress_ms = 0;

        switch (flags_) {
        case G7_FLAG_CTRL_CONNECT_TIMEOUT:
                if (arg_ < 0) {
                        arg_ = 0;
                }
                h->connect_ms = (unsigned int)arg_;
                return pre_return(h, "g7_control", G7_OK, "Connect timeout set to %d ms", h->connect_ms);

        case G7_FLAG_CTRL_PROGRESS_TIMEOUT:
                if (arg_ < 0) {
                        arg_ = 0;
                }
                h->progress_ms = (unsigned int)arg_;
                return pre_return(h, "g7_control", G7_OK, "Progress timeout set to %d ms", h->progress_ms);

        case G7_FLAG_CTRL_SAMPLES_PER_BLOCK:
                // Cannot do this while connected.
                if (h->client != INVALID_SOCKET) {
                        return pre_return(h, "g7_control", G7_ERR_CONNECTED, "Cannot change block size while connected");
                }

                // Range check the block size.
                if ((arg_ < G7SERVER_MIN_BLOCK_SAMPLES) || (arg_ > G7SERVER_MAX_BLOCK_SAMPLES)) {
                        return pre_return(h, "g7_control", G7_ERR_ARGS, "Legal block sizes are %d to %d samples per block", G7SERVER_MIN_BLOCK_SAMPLES, G7SERVER_MAX_BLOCK_SAMPLES);
                }

                // Save locally; will be sent to server at connection time.
                h->samples_per_block = arg_;
                return pre_return(h, "g7_control", G7_OK, "Block size set to %d samples per block", h->samples_per_block);

        case G7_FLAG_CTRL_RATE:
        case G7_FLAG_CTRL_LOOPBACK:
                // Send to server.
                break;

        case G7_FLAG_CTRL_SETUP_C3:
                // Bump up the progress timeout.
                if ((h->progress_ms > 0) && (h->progress_ms < 10000)) {
                        temporary_progress_ms = 10000;
                }

                // Send to server.
                break;

        default:
                return pre_return(h, "g7_control", G7_ERR_ARGS, "Unsupported argument 0x%08X to g7_control", flags_);
        }

        // If we get here, this is a command that goes to the server.
        // Must be connected.
        if (h->client == INVALID_SOCKET) {
                return pre_return(h, "g7_control", G7_ERR_NOTCONNECTED, "Must be connected to server to change this setting (0x%08X)", flags_);
        }

        // Build up a request expecting no data in response.
        memset(&request, 0, sizeof(request));
        request.code = G7_MSG_CONTROL;
        request.arg1 = flags_;
        request.arg2 = arg_;

        if (temporary_progress_ms != 0) {
                saved_progress_ms = h->progress_ms;
                h->progress_ms = temporary_progress_ms;
        }
        // Issue the command.
        ret = issue(h, "g7_control", &request, NULL, NULL, NULL, NULL);
        // Reset the timeout if we modified it.
        if (temporary_progress_ms != 0) {
                h->progress_ms = saved_progress_ms;
        }

        if (ret != G7_OK) {
                // Do not call pre_return; issue() or one of its subordinates will already have done so.
                return ret;
        }

        // Success.
        return pre_return(h, "g7_control", G7_OK, NULL);
}

int g7_connect(g7handle_t handle_, const char* address_)
{
        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;

        char* dupaddress;                               // for address parsing
        char* colon;                                    // for address parsing
        struct sockaddr_in peripheral_addr = { 0 };     // IP address and port of the audio peripheral
        unsigned long yes = 1;                          // value argument for ioctlsocket and setsockopt
        int ret;                                        // general return from system calls
        struct msg_t request;                           // for building the initialize message
        struct msg_t response;                          // Response from the initialize message.
#if !defined(USE_WINSOCK)
        int flags;
#endif

        // If the client socket is already connected, disconnect it.
        if (h->client != INVALID_SOCKET) {
                closesocket(h->client);
                h->client = INVALID_SOCKET;
        }

        // Pre-clear the address.
        memset(&peripheral_addr, 0, sizeof(struct sockaddr_in));

        // NULL address means no connection.
        if ((address_ == NULL) || (*address_ == '\0')) {
                // Disconnect request.
                return pre_return(h, "g7_connect", G7_OK, "Disconnected");
        }

        // Must have a block size in order to connect.
        if (h->samples_per_block == 0) {
                return pre_return(h, "g7_connect", G7_ERR_ARGS, "Must specify block size prior to connecting to a server");
        }

        // Copy the supplied address and break at the colon.
        dupaddress = _strdup(address_);
        colon = strchr(dupaddress, ':');
        if ((colon == NULL) || (strlen(colon + 1) == 0)) {
                free(dupaddress);
                return pre_return(h, "g7_connect", G7_ERR_ARGS, "TCP port must be specified");
        }
        *colon = '\0';

        // Parse the IP address portion into network byte order.
#if defined(USE_WINSOCK)
        peripheral_addr.sin_addr.S_un.S_addr = inet_addr(dupaddress);
        if (peripheral_addr.sin_addr.S_un.S_addr == INADDR_NONE) {
#else
        peripheral_addr.sin_addr.s_addr = inet_addr(dupaddress);
        if (peripheral_addr.sin_addr.s_addr == INADDR_NONE) {
#endif
                free(dupaddress);
                return pre_return(h, "g7_connect", G7_ERR_ARGS, "IP address is not valid");
        }

        // Parse the port portion into network byte order.
        peripheral_addr.sin_port = htons(atoi(colon + 1));
        if (peripheral_addr.sin_port == 0) {
                free(dupaddress);
                return pre_return(h, "g7_connect", G7_ERR_ARGS, "TCP port is not valid");
        }

        // Set the address family.
        peripheral_addr.sin_family = AF_INET;
        free(dupaddress);

        // Create the client socket.
        h->client = socket(AF_INET, SOCK_STREAM, 0);
        if (h->client == INVALID_SOCKET) {
                return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Could not create socket: %s", last_errortext(h));
        }

#if defined(USE_WINSOCK)
        // Set the socket to non-blocking, so we can manage our own timeouts.
        if (ioctlsocket(h->client, FIONBIO, &yes) == SOCKET_ERROR) {
                closesocket(h->client);
                h->client = INVALID_SOCKET;
                return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Could not make socket non-blocking: %s", last_errortext(h));
        }
#else
        // Set the socket to non-blocking.
        flags = fcntl(h->client, F_GETFL);
        if (flags == -1) {
                closesocket(h->client);
                h->client = INVALID_SOCKET;
                return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Could not get socket flags: %s", last_errortext(h));
        }
        flags |= O_NONBLOCK;
        if (fcntl(h->client, F_SETFL, flags) < 0) {
                closesocket(h->client);
                h->client = INVALID_SOCKET;
                return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Could not set socket flags: %s", last_errortext(h));
        }
#endif

        // Disable the Nagle algorithm.
        if (0 != setsockopt(h->client, IPPROTO_TCP, TCP_NODELAY, (const char*)&yes, sizeof(yes))) {
                closesocket(h->client);
                h->client = INVALID_SOCKET;
                return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Could not disable Nagle algorithm: %s", last_errortext(h));
        }

        // Connect to the peripheral.  Normally this will return -1 with a socket error of EWOULDBLOCK,
        // but depending on timing, it is possible for it to return 0 if the connection occurs immediately.
        ret = connect(h->client, (const struct sockaddr*)&peripheral_addr, sizeof(peripheral_addr));
        if (ret == SOCKET_ERROR) {
                int error = last_error();
#if defined(USE_WINSOCK)
                if (error == WSAEWOULDBLOCK) {
#else
                if (error == EINPROGRESS) {
#endif
                        // Normal behavior; the connection has not yet been accepted and we need to wait for it.

                        // Set up for select().
                        struct timeval tv;
                        fd_set writefds;
                        fd_set exceptfds;
                        struct timeval* use_timeout;

                        // Use a timeout if one was specified.
                        use_timeout = NULL;
                        if (h->connect_ms != 0) {
                                tv.tv_sec = (h->connect_ms / 1000);
                                tv.tv_usec = (h->connect_ms - (tv.tv_sec * 1000)) * 1000;
                                use_timeout = &tv;
                        }

                        // Wait for writability or exception.
                        FD_ZERO(&writefds);
                        FD_ZERO(&exceptfds);
#if defined(USE_WINSOCK)
                        writefds.fd_count = 1;
                        writefds.fd_array[0] = h->client;
                        exceptfds.fd_count = 1;
                        exceptfds.fd_array[0] = h->client;
#else
                        FD_SET(h->client, &writefds);
                        FD_SET(h->client, &exceptfds);
#endif

                        // Wait for the socket to go ready for writing, or report an error state (connect failed).
                        // The first parameter (highest socket descriptor in any set) is for POSIX, not Windows.
                        ret = select((int)h->client + 1, NULL, &writefds, &exceptfds, use_timeout);
                        if (ret == SOCKET_ERROR) {
                                closesocket(h->client);
                                h->client = INVALID_SOCKET;
                                return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Error waiting for audio device to connect: %s", last_errortext(h));
                        }
                        if (ret == 0) {
                                // The timeout expired.
                                closesocket(h->client);
                                h->client = INVALID_SOCKET;
                                return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Timed out attempting to connect to audio device");
                        }
                        if (FD_ISSET(h->client, &exceptfds)) {
                                // The connection failed.  Get the error information before we close the socket.
                                const char* error = socket_errortext(h, h->client);
                                closesocket(h->client);
                                h->client = INVALID_SOCKET;
                                return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Error connecting to audio device: %s", error);
                        } else {
                                // The socket is supposedly writable; now check the socket error.
                                int ret;
                                int error;
                                int len;

                                // Get and clear the last-error indication.
                                len = sizeof(error);
                                ret = getsockopt(h->client, SOL_SOCKET, SO_ERROR, (char*)&error, (socklen_t*)&len);
                                if (ret == SOCKET_ERROR) {
                                        // Failed to get the error code itself.
                                        closesocket(h->client);
                                        h->client = INVALID_SOCKET;
                                        return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Could not check SO_ERROR after connect");
                                }
                                if (error != 0) {
                                        closesocket(h->client);
                                        h->client = INVALID_SOCKET;
                                        return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Error after connecting to audio device: %s", errortext(h, error));
                                }
                                // We're good to write.
                        }

                } else {
                        // An error happened in the connect() call itself.
                        closesocket(h->client);
                        h->client = INVALID_SOCKET;
                        return pre_return(h, "g7_connect", G7_ERR_NETWORK, "Error connecting to audio device: %s", last_errortext(h));
                }
        }
        memset(&request, 0, sizeof(request));
        request.code = G7_MSG_INITIALIZE;
        request.arg1 = (PROTOCOL_VERSION_MAJOR << 16) | PROTOCOL_VERSION_MINOR;
        request.arg2 = h->samples_per_block;
        ret = issue(h, "g7_connect", &request, NULL, &response, NULL, NULL);
        if (ret != G7_OK) {
                closesocket(h->client);
                h->client = INVALID_SOCKET;
                return ret;
        }
        DBG_NETWORK(h, g7_errtext(handle_));
        // Connected.  What we're doing here is pulling out the banner that was put in errtext by issue()
        // and letting pre_return stick it back into errtext; a little kludgy but preserves the message.
        return pre_return(h, "g7_connect", G7_OK, g7_errtext(h));
}

int g7_query(g7handle_t handle_, unsigned int flags_, unsigned int* result1_, unsigned int* result2_, unsigned int* result3_, unsigned int* result4_)
{
        struct instance* h = (struct instance*)handle_;
        struct msg_t request;
        struct msg_t response;
        int ret;

        // Result values to be set at the end.
        unsigned int result1 = 0;
        unsigned int result2 = 0;
        unsigned int result3 = 0;
        unsigned int result4 = 0;
        if (h->client == INVALID_SOCKET) {
                return pre_return(h, "g7_query", G7_ERR_NOTCONNECTED, "Must be connected to server to issue queries");
        }

        // Build up a request expecting no data in response, and issue it.
        memset(&request, 0, sizeof(request));
        request.code = G7_MSG_QUERY;
        request.arg1 = flags_;
        ret = issue(h, "g7_query", &request, NULL, &response, NULL, NULL);
        if (ret != G7_OK) {
                return ret;
        }
        result1 = response.arg1;
        result2 = response.arg2;
        result3 = response.arg3;
        result4 = response.arg4;

        // Success.
        if (result1_ != NULL) {
                *result1_ = result1;
        }
        if (result2_ != NULL) {
                *result2_ = result2;
        }
        if (result3_ != NULL) {
                *result3_ = result3;
        }
        if (result4_ != NULL) {
                *result4_ = result4;
        }
        return pre_return(h, "g7_query", G7_OK, "Query 0x%08X returned (0x%08X, 0x%08X, 0x%08X, 0x%08X)", flags_, result1, result2, result3, result4);
}

int g7_stage_output(g7handle_t handle_, unsigned int flags_, unsigned int channels_, const sample_t** data_, unsigned int repeat_, unsigned int tag_)
{
        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;

        struct msg_t request;
        int ret = G7_OK;
        int appending = 0;
        unsigned int channel_bitmap = 0;
        unsigned int send_channels = 0;
        sample_t* cloned_data = NULL;
        sample_t* data_in_clone_buffer = NULL;
        unsigned int channel;
        unsigned int sample;

        appending = (flags_ & (G7_FLAG_SO_APPEND_DATA | G7_FLAG_SO_APPEND_RAMP_UP | G7_FLAG_SO_APPEND_RAMP_DOWN));
        if (appending) {
                if (data_ != NULL) {
                        for (channel = 0; channel < channels_; channel++) {
                                if (data_[channel] != NULL) {
                                        channel_bitmap |= (1 << channel);
                                        send_channels++;
                                }
                        }
                }

                // Allocate local storage to hold the non-NULL channels, and clone data into it.
                if (send_channels != 0) {

                        // Allocate a buffer big enough to hold the non-NULL blocks one after the other.
                        cloned_data = (sample_t*)malloc(sizeof(sample_t) * send_channels * h->samples_per_block);

                        // Clone user-supplied data into the big clone buffer.
                        data_in_clone_buffer = cloned_data;
                        for (channel = 0; channel < G7SERVER_MAXIMUM_CHANNELS; channel++) {

                                // Skip channels which the caller doesn't want.
                                if (0 == (channel_bitmap & (1 << channel))) {
                                        continue;
                                }

                                // Copy the caller-supplied data into the clone buffer, one buffer after the other.
                                memcpy(data_in_clone_buffer, data_[channel], sizeof(sample_t) * h->samples_per_block);

                                // If they want a ramp-up or ramp-down block instead of the caller-supplied data, convert the
                                // cloned data to the appropriate kind of ramp.  (Don't need to ramp zeroes--they're zeroes).
                                if (flags_ & (G7_FLAG_SO_APPEND_RAMP_UP | G7_FLAG_SO_APPEND_RAMP_DOWN)) {
                                        double step;
                                        double scale;

                                        // Ramp either up or down.
                                        step = 1.0 / (double)h->samples_per_block;
                                        scale = 0.0;
                                        for (sample = 0; sample < h->samples_per_block; sample++) {
                                                if (flags_ & G7_FLAG_SO_APPEND_RAMP_UP) {
                                                        data_in_clone_buffer[sample] = (sample_t)(data_in_clone_buffer[sample] * scale);
                                                } else {
                                                        data_in_clone_buffer[h->samples_per_block - 1 - sample] = (sample_t)(data_in_clone_buffer[h->samples_per_block - 1 - sample] * scale);
                                                }
                                                scale += step;
                                                if (scale > 1.0) {
                                                        scale = 1.0;
                                                }
                                        }
                                }

                                // Set up to copy the next non-NULL source data into the clone buffer immediately after the block we just did.
                                data_in_clone_buffer += h->samples_per_block;
                        }
                }

                // Convert the flag to just G7_FLAG_SO_APPEND_DATA for the server (it doesn't do pattern generation)
                flags_ &= ~(G7_FLAG_SO_APPEND_RAMP_UP | G7_FLAG_SO_APPEND_RAMP_DOWN);
                flags_ |= G7_FLAG_SO_APPEND_DATA;
        }

        // Send the data to the server.
        // If we need to do byte reversal, do it here on everything in the cloned block.
        // Note that channel here isn't really the channel; it's the channel relative to non-NULL channels.
        if (htonl(0x12345678) != 0x12345678) {
                unsigned int samples = send_channels * h->samples_per_block;
                for (sample = 0; sample < samples; sample++) {
                        cloned_data[sample] = htonl(cloned_data[sample]);
                }
        }

        // Send a request to the audio device, expecting no data in return.
        // arg1 = flags
        // arg2 = with a G7_FLAG_SO_APPEND_ flag, bitmap of which channels are non-NULL; MSB = channel 0
        // arg3 = with a G7_FLAG_SO_APPEND_ flag, repeat count; otherwise not used
        // arg4 = with a G7_FLAG_SO_APPEND_ flag, tag; otherwise not used
        // data = with a G7_FLAG_SO_APPEND_ flag, bitcnt(arg2) channels worth of data; otherwise not used
        memset(&request, 0, sizeof(request));
        request.code = G7_MSG_STAGE_OUTPUT;
        request.arg1 = flags_;
        request.arg2 = channel_bitmap;
        request.arg3 = repeat_;
        request.arg4 = tag_;
        request.following_bytes = send_channels * h->samples_per_block * sizeof(sample_t);

        ret = issue(h, "g7_stage_output", &request, (const char*)cloned_data, NULL, NULL, NULL);

        // Clean up the monolithic chunk of cloned data.
        if (cloned_data != NULL) {
                free(cloned_data);
        }

        // Don't call pre_return on error; issue() or one of its subordinates will already have done so.
        if (ret != G7_OK) {
                return ret;
        }
        return pre_return(h, "g7_stage_output", G7_OK, NULL);
}

int g7_control_output(g7handle_t handle_, unsigned int flags_)
{
        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;

        struct msg_t request;
        struct msg_t response;
        int ret;

        // Error if no block size.
        if (h->samples_per_block == 0) {
                return pre_return(h, "g7_control_output", G7_ERR_STATE, "Cannot set data before setting block size");
        }

        // Build up a request expecting no data in response, and issue it.
        memset(&request, 0, sizeof(request));
        request.code = G7_MSG_CONTROL_OUTPUT;
        request.arg1 = flags_;
        ret = issue(h, "g7_control_output", &request, NULL, &response, NULL, NULL);
        if (ret != G7_OK) {
                // Note that if an error occurred, we don't necessarily know what state the server is in; for example the request
                // could have been lost, or the response could have been lost.  So we don't change our state in response to a bad
                // return; the caller can always call us with G7_FLAG_CO_STOP again to try to get back into a legitimate state.

                // Do not call pre_return; issue() or one of its subordinates will already have done so.
                return ret;
        }

        // Success.
        return pre_return(h, "g7_control_output", G7_OK, "Operation with flags 0x%08X succeeded", flags_);
}

int g7_read_input(g7handle_t handle_, unsigned int channels_, sample_t* buffer_, unsigned int* blocks_, unsigned int flags_, unsigned int* tag_, unsigned int* sequence_)
{
        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;

        struct msg_t request;
        struct msg_t response;
        int ret;
        unsigned int data_bytes;
        unsigned int requested_blocks;
        unsigned int requested_channels;
        unsigned int sample;
        sample_t* dst;
        unsigned int channel;
        unsigned int reported_tag = 0;
        unsigned int reported_sequence = 0;
        int byte_swap;
        int right_justify;

        // Must specify the block count (used for both in and out), and it must not be zero.
        if (blocks_ == NULL) {
                return pre_return(h, "g7_read_input", G7_ERR_ARGS, "Must specify block count");
        }
        if (*blocks_ == 0) {
                return pre_return(h, "g7_read_input", G7_ERR_ARGS, "Must specify non-zero block count");
        }
        if (buffer_ == NULL) {
                return pre_return(h, "g7_read_input", G7_ERR_ARGS, "Result buffer pointer cannot be NULL");
        }

        // Remember the requested block count, and preset the returned block count to zero.
        requested_blocks = *blocks_;
        *blocks_ = 0;

        // Where is data to be returned?
        dst = buffer_;

        // Count the maximum size of the return buffer.
        requested_channels = 0;
        for (channel = 0; channel < G7SERVER_MAXIMUM_CHANNELS; channel++) {
                if (channels_ & (1 << channel)) {
                        requested_channels++;
                }
        }
        data_bytes = requested_channels * requested_blocks * h->samples_per_block * sizeof(int);

        // Build up a request expecting data in response, and issue it.
        memset(&request, 0, sizeof(request));
        request.code = G7_MSG_READ_INPUT;
        request.arg1 = channels_;
        request.arg2 = requested_blocks;
        // Note that G7_FLAG_24_RJ_IN_32 is handled in the client, not the server.
        request.arg3 = flags_ & (G7_FLAG_READ_DISCARD | G7_FLAG_READ_LATEST);
        ret = issue(h, "g7_read_input", &request, NULL, &response, (char*)buffer_, &data_bytes);
        if (ret != G7_OK) {
                // Don't call pre_return; issue() or one of its subordinates will already have done so.
                return ret;
        }

        // The following_bytes field in the response will tell us how much data we actually got.
        // According the protocol, the peripheral must send us an integral number of blocks.
        // It's normal to receive zero blocks either at end-of-stream, or if we are polling
        // for data faster than the capture operation is proceeding at the peripheral end.
        else if ((data_bytes % (requested_channels * h->samples_per_block * sizeof(sample_t))) != 0) {
                return pre_return(h, "g7_read_input", G7_ERR_NETWORK, "Received %d bytes of data, which is not a multiple of %d x block size (%d) x channels (%d)",
                        data_bytes, sizeof(sample_t), h->samples_per_block, requested_channels);
        }

        // If we received no data, check to see if we are at the end of the stream; if so, we can
        // report -1 to tell the caller that they don't need to keep polling this channel for data.
        if ((data_bytes == 0) && (response.arg1 == 1)) {
                return pre_return(h, "g7_read_input", G7_ERR_END_OF_STREAM, "End-of-stream indication from server");
        }

        // Report the number of blocks we received.buffer_
        *blocks_ = data_bytes / (requested_channels * h->samples_per_block * sizeof(sample_t));

        // Report other values of interest.
        reported_tag = response.arg2;
        reported_sequence = response.arg3;

        // Do we need to byte-swap or right-justify the received data?
        // NOTE: there is an assumption here that sizeof(sample_t) == 4.
        byte_swap = (ntohl(0x12345678) != 0x12345678);
        right_justify = (flags_ & G7_FLAG_READ_24_RJ_IN_32);
        if (byte_swap || right_justify) {

                // Process the data in-place in the caller-provided buffer.
                unsigned int samples = data_bytes / sizeof(sample_t);
                dst = buffer_;

                // Do both operations at once if we can.
                // WARNING: >> is implementation-dependent as to whether it sign extends.  So we sign extend manually.
                if (byte_swap && right_justify) {
                        for (sample = 0; sample < samples; sample++) {
                                sample_t v = ntohl(*dst);
                                v = (v >> 8) | (v < 0 ? 0xFF000000 : 0);
                                *dst = v;
                                dst++;
                        }
                } else if (right_justify) {
                        for (sample = 0; sample < samples; sample++) {
                                *dst = (*dst >> 8) | (*dst < 0 ? 0xFF000000 : 0);
                                dst++;
                        }
                } else if (byte_swap) {
                        for (sample = 0; sample < samples; sample++) {
                                *dst = ntohl(*dst);
                                dst++;
                        }
                }
        }

        // Put tag and sequence in user-provided result fields.  We can't get here without having set
        // reported_tag and reported_sequence, since we will already have returned if we got no data.
        if (tag_ != NULL) {
                *tag_ = reported_tag;
        }
        if (sequence_ != NULL) {
                *sequence_ = reported_sequence;
        }

        // Success.  We can use blocks_ here unconditionally because we already made sure it wasn't NULL.
        return pre_return(h, "g7_read_input", G7_OK, "Returned %d blocks (tag 0x%08X, sequence %d)", *blocks_, reported_tag, reported_sequence);
}

int g7_link(g7handle_t handle_, unsigned char const *command_, unsigned int byte_count_, unsigned char* response_)
{
        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;

        struct msg_t request;
        struct msg_t response;
        int ret;
        unsigned int num_bytes;

        if (command_ == NULL) {
                return pre_return(h, "g7_link", G7_ERR_ARGS, "Command buffer pointer cannot be NULL.");
        }
        if (*command_ == 0) {
                return pre_return(h, "g7_link", G7_ERR_ARGS, "First command byte must be non-zero.");
        }
        if (response_ == NULL) {
                return pre_return(h, "g7_link", G7_ERR_ARGS, "Response buffer pointer cannot be NULL");
        }

        // Build up a request that contains data, and expects data in the response, and issue it.
        memset(&request, 0, sizeof(request));
        request.code = G7_MSG_LINK;
        request.following_bytes = num_bytes = byte_count_;

        ret = issue(h, "g7_link", &request, (const char*)command_, &response, (char*)response_, &num_bytes);
        if (ret != G7_OK) {
                // Don't call pre_return; issue() or one of its subordinates will already have done so.
                return ret;
        }

        // The following_bytes field in the response will tell us how much data we actually got.
        else if (num_bytes != byte_count_) {
                return pre_return(h, "g7_link", G7_ERR_NETWORK, "Received %d bytes of data, but expected %d bytes", num_bytes, byte_count_);
        }

        return pre_return(h, "g7_link", G7_OK, "Received %d-byte response", num_bytes, byte_count_);
}

void g7_shutdown(g7handle_t handle_)
{
        // Recover the instance pointer.
        struct instance* h = (struct instance*)handle_;
        if (h == NULL) {
                // No effect if a NULL handle was specified.
                return;
        }

        // Don't issue a stop command; the server should time out anyway.

        // No return value, but we can log the call anyway.
        pre_return(h, "g7_shutdown", G7_OK, NULL);

        // Free the instance.
        free(h);

        // Keep track of the number of active instances.
        active_instances--;

#if defined(USE_WINSOCK)
        // If we were the last instance, shut down Windows Sockets.
        if (active_instances == 0) {
                WSACleanup();
        }
#endif
}
