
//Subversion Repositories repos 1
//(root)/Gen7/HandheldSourceCode/Overlay/recipes/g7play/files/g7play.cpp - Rev 13706
//Rev 13020 | Blame | Compare with Previous | Last modification | View Log | RSS feed
// START

// Gen7 Audio Test Client
// Copyright (c) 2010-2015 Mimosa Acoustics, Inc.

// Linux end-of-line markers, 8-column _tabs_ (not spaces), K&R C brace style.

// For Dev Studio 2005, Had to add an Additional Dependency to the linker
// go to Project-> <Project Name> Properties -> Configuration Properties -> Linker ->
// Input and add Ws2_32.lib to the Additional Dependencies.

// The version string to display.
#define G7PLAY_VERSION  "2018.03.30"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_MSC_VER)
#include <windows.h>            // for Sleep, GetTickCount
// The POSIX names are used in the code; #define them to the non-standard Windows equivalents.
#define strcasecmp _stricmp
#define strncasecmp _strnicmp

#define ssize_t _off_t
#define MAXPATHLEN MAX_PATH

// Suppress complaints that we're not using Microsoft's non-standard _s versions of string handling functions.
// Defining _CRT_SECURE_NO_WARNINGS is the advertised solution, but doesn't work in all cases, thus the pragma.
#define _CRT_SECURE_NO_WARNINGS 1
#pragma warning(disable: 4996)

// Link with the Windows Sockets library.
#pragma comment(lib, "wsock32.lib")
#else
#include <unistd.h>
#include <sys/param.h>          // for MAXPATHLEN
#include <sys/time.h>
#endif

#include <errno.h>
#include <sys/stat.h>

#define _USE_MATH_DEFINES
#include <math.h>               // sin(), M_PI

//  Windows can't deal with symbolic links, so we have to know where to find these files.
#if defined(_MSC_VER)
#include "../../g7common/g7client.h"
#else
#include "g7client.h"
#endif

// Sample rate (hard coded at 48kHz for now)
#define SAMPLES_PER_SECOND      48000

// Blocks to read at one time.
#define BLOCKS_PER_READ         10

// The number of header (quiet or loopback) samples which g7server needs to ensure that the I2S loopback is down.
#define SYNC_SAMPLES            4000

// The number of blocks after the ramp-up block (or at the start of the sequence if no ramp-up block is used) to discard by default.
#define DEFAULT_JUNK_BLOCKS     1

// Mechanism to map loopback values to names, and vice versa.
static const struct loopback_t {
        const char* tag;
        int val;
        const char* desc;
} loopback_levels[] =
{
        { "none",               G7_LOOPBACK_NONE,               "No loopback" },
        { "server",             G7_LOOPBACK_SERVER,             "Loopback within server software" },
        { "i2s",                G7_LOOPBACK_I2S,                "Loopback on 1-bit digital path to codec" }
};
static const unsigned int num_loopback_levels = sizeof(loopback_levels) / sizeof(struct loopback_t);

// Look up a loopback entry by value.  Returns a non-NULL value or exits.
static const struct loopback_t* loopback_by_value(int loopback_)
{
        unsigned int level;
        for (level = 0; level < num_loopback_levels; level++) {
                if (loopback_ == loopback_levels[level].val) {
                        return &loopback_levels[level];
                }
        }
        printf("Unrecognized loopback value %d.\n", loopback_);
        exit(1);
}

// Look up a loopback entry by tag.  Returns a non-NULL value or exits.
static const struct loopback_t* loopback_by_tag(const char* tag_)
{
        unsigned int level;
        for (level = 0; level < num_loopback_levels; level++) {
                if (strcasecmp(tag_, loopback_levels[level].tag) == 0) {
                        return &loopback_levels[level];
                }
        }
        printf("Unrecognized loopback \"%s\".\n", tag_);
        exit(1);
}

// For sine wave monitoring.

// The number of successive zero crossings we examine.
#define NUM_ZERO_CROSSINGS      4

// The number of samples we have previously received.
static int absolute_sample;

// The time of the last rising zero crossings.
static int rising_zero[NUM_ZERO_CROSSINGS];

// The next-in index of the next rising zero.
static int next_rising_zero;

// The maximum signed sample value seen.
static sample_t highest_value;

// The minimum negative sample value seen.
static sample_t lowest_value;

// The previous block of samples.
static sample_t* previous_block;

// The current block of samples.
static sample_t* current_block;

#if defined(_MSC_VER)
// Get the millisecond tick count
static unsigned int tick(void)
{
        return ::GetTickCount();
}
#else
// The time when tick() was first called.
static struct timeval first_tick_call = { 0, 0 };

// Get the millisecond tick count
static unsigned int tick(void)
{
        // Get the current time.
        struct timeval now;
        gettimeofday(&now, NULL);

        // If this is the first time report, save it.
        if (first_tick_call.tv_sec == 0) {
                first_tick_call = now;
        }

        // Report the difference in milliseconds.
        return ((now.tv_sec - first_tick_call.tv_sec) * 1000) + ((now.tv_usec - first_tick_call.tv_usec) / 1000);
}
#endif

static void cleanup_follower(void)
{
        if (previous_block != NULL) {
                free(previous_block);
                previous_block = NULL;
        }
}

static void setup_follower(void)
{
        absolute_sample = 0;
        next_rising_zero = 0;
        highest_value = 0x80000000;
        lowest_value = 0x7FFFFFFF;
        previous_block = NULL;
        current_block = NULL;

        for (int zero = 0; zero < NUM_ZERO_CROSSINGS; zero++) {
                rising_zero[zero] = 0;
        }
}

static void analyze_block(int samples_per_block_, const sample_t* current_block_)
{
        //printf("analyze block, %d samples, abs sample = %d\n", samples_per_block_, absolute_sample);
        // Look for zero crossings.
        for (int sample = 0; sample < (int)samples_per_block_ - 3; sample++) {
                int cm3 = ((sample - 3) < 0) ? ((previous_block == NULL) ? 0 : previous_block[samples_per_block_ - 1 - 3]) : current_block_[sample - 3];
                int cm2 = ((sample - 2) < 0) ? ((previous_block == NULL) ? 0 : previous_block[samples_per_block_ - 1 - 2]) : current_block_[sample - 2];
                int cm1 = ((sample - 1) < 0) ? ((previous_block == NULL) ? 0 : previous_block[samples_per_block_ - 1 - 1]) : current_block_[sample - 1];
                sample_t c0 = current_block_[sample];
                sample_t cp1 = current_block_[sample + 1];
                sample_t cp2 = current_block_[sample + 2];

                // Track the highest and lowest values.
                if ((absolute_sample == 0) || (c0 > highest_value)) {
                        highest_value = c0;
                }
                if ((absolute_sample == 0) || (c0 < lowest_value)) {
                        lowest_value = c0;
                }

                // Advance the absolute sample count.
                absolute_sample++;

                // Is there a rising zero crossing between cm1 and c0?
                if ((cm3 < 0) && (cm2 < 0) && (cm1 < 0) && (c0 >= 0) && (cp1 >= 0) && (cp2 >= 0)) {
                        //fprintf(ci.file, "z %10d %10d\n", sample + ci.absolute_sample, c0);

                        // Yes.
                        rising_zero[next_rising_zero] = absolute_sample;
                        //printf("rising zero [%d] at sample %d\n", next_rising_zero, absolute_sample);

                        // Advance next_rising_zero to point to the oldest remaining zero in the buffer.
                        next_rising_zero = (next_rising_zero + 1) % NUM_ZERO_CROSSINGS;

                        // The shortest and longest interval between successive zeroes.
                        // (preinitialized to force replacement during min-maxing).
                        int shortest_interval  = (1 << 30);
                        int longest_interval = -1;

                        // The total of all intervals and the number of valid intervals, for computing a moving average.
                        int total_interval = 0;
                        int num_intervals = 0;

                        // next_rising_zero now points to the oldest zero in the ring.
                        // Walk through each of the NUM_ZERO_CROSSING intervals in the ring.
                        for (unsigned int interval = 0; interval < NUM_ZERO_CROSSINGS - 1; interval++) {

                                // Compute the positions in the ring of the two comparison targets.
                                unsigned int older_zero_index = (next_rising_zero + interval) % NUM_ZERO_CROSSINGS;
                                unsigned int newer_zero_index = (next_rising_zero + interval + 1) % NUM_ZERO_CROSSINGS;

                                // Get the sample numbers during which the zeroes occurred.
                                int older_zero_sample = rising_zero[older_zero_index];
                                int newer_zero_sample = rising_zero[newer_zero_index];

                                // Until the ring fills up, some samples will be zero.
                                if (older_zero_sample > 0) {
                                        int span = newer_zero_sample - older_zero_sample;
                                        if (span < shortest_interval) {
                                                shortest_interval = span;
                                        }
                                        if (span > longest_interval) {
                                                longest_interval = span;
                                        }
                                        total_interval += span;
                                        num_intervals++;
                                }
                        }

                        // Compute the percentage of full scale.
                        unsigned int full_range = highest_value - lowest_value;
                        int range_percent = (full_range >> 24) * 100 / 255;

                        // If we have a full ring and there is less than 5 samples worth of jitter, we know the frequency.
                        if ((num_intervals == (NUM_ZERO_CROSSINGS - 1)) && ((longest_interval - shortest_interval) < 5)) {

                                // Compute the average frequency.
                                double samples_per_cycle  = (double)total_interval / (double)num_intervals;
                                double seconds_per_cycle = samples_per_cycle / 48000.0;
                                double frequency = 1.0 / seconds_per_cycle;
                                printf("input locked at %f Hz, %d%% of full scale (0x%08X to 0x%08X)\n", frequency, range_percent, lowest_value, highest_value);
                        } else if (num_intervals != 0) {

                                // Ring not yet full, or no stable frequency.
                                printf("input unlocked (%d intervals, %d..%d), %d%% of full scale (0x%08X to 0x%08X)\n",
                                        num_intervals, shortest_interval, longest_interval, range_percent, lowest_value, highest_value);
                        }
                }
        }

        // Make the current block the new previous block.
        if (previous_block == NULL) {
                previous_block = (sample_t*)malloc(samples_per_block_ * sizeof(sample_t));
        }
        memcpy(previous_block, current_block_, samples_per_block_ * sizeof(sample_t));
}

// Rev C3: Convert a string of one or more hardware settings to a bit mask and a data word.
// s_ is the string to be interpreted.
// The bits in the word to be passed to g7_control(G7_FLAG_CTRL_SETUP_C3) is stored at *data_.
// The mask of relevant bits to be changed in the state word is stored at *mask_.
// (Fields which are not referenced in s_ will not have their correponding bits set in *mask_).
// Returns 1 on error, 0 on success.
unsigned int parse_hardware_settings(const char* s_, unsigned int* data_, unsigned int* mask_)
{
        unsigned int mask = 0;
        unsigned int data = 0;

        while (*s_ != '\0') {
                if ((*s_ == ',') || (*s_ == ' ')) {
                        s_++;
                }

                // Differential probe power
                else if (strncasecmp(s_, "ppon", 4) == 0) {
                        mask |= G7_SETUP_C3_PROBE_POWER_MASK;
                        data |= G7_SETUP_C3_PROBE_POWER_ON;
                        s_ += 4;
                } else if (strncasecmp(s_, "ppoff", 5) == 0) {
                        mask |= G7_SETUP_C3_PROBE_POWER_MASK;
                        data |= G7_SETUP_C3_PROBE_POWER_OFF;
                        s_ += 5;
                }

                // Single-ended microphone bias
                else if (strncasecmp(s_, "mbon", 4) == 0) {
                        mask |= G7_SETUP_C3_MIC_BIAS_MASK;
                        data |= G7_SETUP_C3_MIC_BIAS_ON;
                        s_ += 4;
                } else if (strncasecmp(s_, "mboff", 5) == 0) {
                        mask |= G7_SETUP_C3_MIC_BIAS_MASK;
                        data |= G7_SETUP_C3_MIC_BIAS_OFF;
                        s_ += 5;
                }

                // Single-ended microphone preamp
                else if (strncasecmp(s_, "mpon", 4) == 0) {
                        mask |= G7_SETUP_C3_MIC_PREAMP_MASK;
                        data |= G7_SETUP_C3_MIC_PREAMP_ON;
                        s_ += 4;
                } else if (strncasecmp(s_, "mpoff", 5) == 0) {
                        mask |= G7_SETUP_C3_MIC_PREAMP_MASK;
                        data |= G7_SETUP_C3_MIC_PREAMP_OFF;
                        s_ += 5;
                }

                // Primary input
                else if (strncasecmp(s_, "moff", 4) == 0) {
                        mask |= G7_SETUP_C3_PRIMARY_INPUT_MASK;
                        data |= G7_SETUP_C3_PRIMARY_INPUT_OFF;
                        s_ += 4;
                } else if (strncasecmp(s_, "mlo", 3) == 0) {
                        mask |= G7_SETUP_C3_PRIMARY_INPUT_MASK;
                        data |= G7_SETUP_C3_PRIMARY_INPUT_LOW;
                        s_ += 3;
                } else if (strncasecmp(s_, "mhi", 3) == 0) {
                        mask |= G7_SETUP_C3_PRIMARY_INPUT_MASK;
                        data |= G7_SETUP_C3_PRIMARY_INPUT_HIGH;
                        s_ += 3;
                }
                // New values for Rev E
                else if (strncasecmp(s_, "mfm", 3) == 0) {
                        mask |= G7_SETUP_C3_PRIMARY_INPUT_MASK;
                        data |= G7_SETUP_C3_E_PRIMARY_INPUT_M;
                        s_ += 3;
                } else if (strncasecmp(s_, "mfma", 3) == 0) {
                        mask |= G7_SETUP_C3_PRIMARY_INPUT_MASK;
                        data |= G7_SETUP_C3_E_PRIMARY_INPUT_MA;
                        s_ += 4;
                } else if (strncasecmp(s_, "mfmb", 3) == 0) {
                        mask |= G7_SETUP_C3_PRIMARY_INPUT_MASK;
                        data |= G7_SETUP_C3_E_PRIMARY_INPUT_MB;
                        s_ += 4;
                } else if (strncasecmp(s_, "mfmab", 3) == 0) {
                        mask |= G7_SETUP_C3_PRIMARY_INPUT_MASK;
                        data |= G7_SETUP_C3_E_PRIMARY_INPUT_MAB;
                        s_ += 5;
                }

                // Auxiliary input
                else if (strncasecmp(s_, "auxoff", 6) == 0) {
                        mask |= G7_SETUP_C3_AUX_INPUT_MASK;
                        data |= G7_SETUP_C3_AUX_INPUT_OFF;
                        s_ += 6;
                } else if (strncasecmp(s_, "auxe", 4) == 0) {
                        mask |= G7_SETUP_C3_AUX_INPUT_MASK;
                        data |= G7_SETUP_C3_AUX_INPUT_EXTERNAL;
                        s_ += 4;
                } else if (strncasecmp(s_, "auxa", 4) == 0) {
                        mask |= G7_SETUP_C3_AUX_INPUT_MASK;
                        data |= G7_SETUP_C3_AUX_INPUT_MON_A;
                        s_ += 4;
                } else if (strncasecmp(s_, "auxb", 4) == 0) {
                        mask |= G7_SETUP_C3_AUX_INPUT_MASK;
                        data |= G7_SETUP_C3_AUX_INPUT_MON_B;
                        s_ += 4;
                }

                // Receiver A
                else if (strncasecmp(s_, "aoff", 4) == 0) {
                        mask |= G7_SETUP_C3_OUT_A_MASK;
                        data |= G7_SETUP_C3_OUT_A_OFF;
                        s_ += 4;
                } else if (strncasecmp(s_, "alo", 3) == 0) {
                        mask |= G7_SETUP_C3_OUT_A_MASK;
                        data |= G7_SETUP_C3_OUT_A_ATTENUATED;
                        s_ += 3;
                } else if (strncasecmp(s_, "ahi", 3) == 0) {
                        mask |= G7_SETUP_C3_OUT_A_MASK;
                        data |= G7_SETUP_C3_OUT_A_UNATTENUATED;
                        s_ += 3;
                }

                // Receiver B
                else if (strncasecmp(s_, "boff", 4) == 0) {
                        mask |= G7_SETUP_C3_OUT_B_MASK;
                        data |= G7_SETUP_C3_OUT_B_OFF;
                        s_ += 4;
                } else if (strncasecmp(s_, "blo", 3) == 0) {
                        mask |= G7_SETUP_C3_OUT_B_MASK;
                        data |= G7_SETUP_C3_OUT_B_ATTENUATED;
                        s_ += 3;
                } else if (strncasecmp(s_, "bhi", 3) == 0) {
                        mask |= G7_SETUP_C3_OUT_B_MASK;
                        data |= G7_SETUP_C3_OUT_B_UNATTENUATED;
                        s_ += 3;
                }

                // Unrecognized keyword
                else {
                        // Error.
                        return 1;
                }
        }

        // Save results.
        if (data_ != NULL) {
                *data_ = data;
        }
        if (mask_ != NULL) {
                *mask_ = mask;
        }

        // Success.
        return 0;
}

// Rev C3: print a description of bits in the state word returned by g7_query(G7_FLAG_QUERY_SETUP).
// state_ is the state word; mask_ is the bitmask of fields in which we are interested.
void interpret_hardware_settings(unsigned int state_, unsigned int mask_)
{
        const char* comma = "";

        // Master state
        if (mask_ & G7_SETUP_C3_MASTER_MASK) {
                switch (state_ & G7_SETUP_C3_MASTER_MASK) {
                case G7_SETUP_C3_MASTER_OFF:
                        printf("off");
                        break;
                case G7_SETUP_C3_MASTER_STANDBY:
                        printf("standby");
                        break;
                case G7_SETUP_C3_MASTER_16KHZ:
                        printf("running at 16KHz");
                        break;
                case G7_SETUP_C3_MASTER_24KHZ:
                        printf("running at 24KHz");
                        break;
                case G7_SETUP_C3_MASTER_32KHZ:
                        printf("running at 32KHz");
                        break;
                case G7_SETUP_C3_MASTER_48KHZ:
                        printf("running at 48KHz");
                        break;
                }
                comma = "; ";
        }

        // Differential probe power
        if (mask_ & G7_SETUP_C3_PROBE_POWER_MASK) {
                printf("%s%s", comma, state_ & G7_SETUP_C3_PROBE_POWER_ON ? "ppon" : "ppoff");
                comma = ",";
        }

        // Single-ended microphone bias
        if (mask_ & G7_SETUP_C3_MIC_BIAS_MASK) {
                printf("%s%s", comma, state_ & G7_SETUP_C3_MIC_BIAS_ON ? "mbon" : "mboff");
                comma = ",";
        }

        // Single-ended microphone preamp
        if (mask_ & G7_SETUP_C3_MIC_PREAMP_MASK) {
                printf("%s%s", comma, state_ & G7_SETUP_C3_MIC_PREAMP_ON ? "mpon" : "mpoff");
                comma = ",";
        }

        // Primary input
        if (mask_ & G7_SETUP_C3_PRIMARY_INPUT_MASK) {
                switch (state_ & G7_SETUP_C3_PRIMARY_INPUT_MASK) {
                case G7_SETUP_C3_PRIMARY_INPUT_OFF:
                        printf("%smoff", comma);
                        break;
                case G7_SETUP_C3_PRIMARY_INPUT_LOW:
                        printf("%smlo", comma);
                        break;
                case G7_SETUP_C3_PRIMARY_INPUT_HIGH:
                        printf("%smhi", comma);
                        break;
                // New values for Rev E.
                case G7_SETUP_C3_E_PRIMARY_INPUT_M:
                        printf("%smfm", comma);
                        break;
                case G7_SETUP_C3_E_PRIMARY_INPUT_MA:
                        printf("%smfma", comma);
                        break;
                case G7_SETUP_C3_E_PRIMARY_INPUT_MB:
                        printf("%smfmb", comma);
                        break;
                case G7_SETUP_C3_E_PRIMARY_INPUT_MAB:
                        printf("%smfmab", comma);
                        break;
                default:
                        printf("%sm?!", comma);
                        break;
                }
                comma = ",";
        }

        // Auxiliary input
        if (mask_ & G7_SETUP_C3_AUX_INPUT_MASK) {
                switch (state_ & G7_SETUP_C3_AUX_INPUT_MASK) {
                case G7_SETUP_C3_AUX_INPUT_OFF:
                        printf("%sauxoff", comma);
                        break;
                case G7_SETUP_C3_AUX_INPUT_EXTERNAL:
                        printf("%sauxe", comma);
                        break;
                case G7_SETUP_C3_AUX_INPUT_MON_A:
                        printf("%sauxa", comma);
                        break;
                case G7_SETUP_C3_AUX_INPUT_MON_B:
                        printf("%sauxb", comma);
                        break;
                default:
                        printf("%saux?!", comma);
                        break;
                }
                comma = ",";
        }

        // Receiver A
        if (mask_ & G7_SETUP_C3_OUT_A_MASK) {
                switch (state_ & G7_SETUP_C3_OUT_A_MASK) {
                case G7_SETUP_C3_OUT_A_OFF:
                        printf("%saoff", comma);
                        break;
                case G7_SETUP_C3_OUT_A_ATTENUATED:
                        printf("%salo", comma);
                        break;
                case G7_SETUP_C3_OUT_A_UNATTENUATED:
                        printf("%sahi", comma);
                        break;
                default:
                        printf("%sa?!", comma);
                        break;
                }
                comma = ",";
        }

        // Receiver B
        if (mask_ & G7_SETUP_C3_OUT_B_MASK) {
                switch (state_ & G7_SETUP_C3_OUT_B_MASK) {
                case G7_SETUP_C3_OUT_B_OFF:
                        printf("%sboff", comma);
                        break;
                case G7_SETUP_C3_OUT_B_ATTENUATED:
                        printf("%sblo", comma);
                        break;
                case G7_SETUP_C3_OUT_B_UNATTENUATED:
                        printf("%sbhi", comma);
                        break;
                default:
                        printf("%sb?!", comma);
                        break;
                }
                comma = ",";
        }
}

// Default argument values.
#define DEFAULT_DATA_BLOCKS                     100
#define DEFAULT_FRONT_QUIET_BLOCKS              10
#define DEFAULT_END_QUIET_BLOCKS                1
#define DEFAULT_SAMPLES_PER_BLOCK               1024
#define DEFAULT_FREQUENCY                       440
#define DEFAULT_ADDRESS                         G7SERVER_DEFAULT_ADDRESS
#define DEFAULT_LOOPBACK                        G7_LOOPBACK_NONE
#define DEFAULT_SAMPLE_RATE                     48000

int main(int argc, char* argv[])
{
        // Start with defaults.
        int loopback = DEFAULT_LOOPBACK;
        unsigned int data_blocks = DEFAULT_DATA_BLOCKS;
        unsigned int front_quiet_blocks = DEFAULT_FRONT_QUIET_BLOCKS;
        unsigned int end_quiet_blocks = DEFAULT_END_QUIET_BLOCKS;
        unsigned int samples_per_block = DEFAULT_SAMPLES_PER_BLOCK;
        const char* address = DEFAULT_ADDRESS;
        int frequency = DEFAULT_FREQUENCY;
        int test_pattern = 0;
        int test_level = 0;
        int stop_ms = 0;
        int ramp_blocks = 0;
        int silence = 0;
        int promiscuous = 0;
        unsigned int sample_rate = DEFAULT_SAMPLE_RATE;
        unsigned int g7client_debug_mask = 0;
        const char* g7client_log_path = NULL;
        // extended = 0 if not an extended sequence, 1 for -x, 2 for -xl
        int extended = 0;
        int extended_blocks = 0;
        int junk_blocks = DEFAULT_JUNK_BLOCKS;
        unsigned int relay_set = 0;
        const char* oc_cmd = NULL;
        unsigned int new_state_mask = 0;
        unsigned int new_state_data = 0;
        int sync_sample_check = 1;
        int synchronize = 1;

        // Sine wave parameters and sample buffer.
        double samples_per_second = SAMPLES_PER_SECOND;
        double seconds_per_block;
        double periods_per_second;
        double periods_per_block;
        sample_t* sine;
        double amplitude = 1.0;

        // Test pattern buffer.
        sample_t* test;

        // Set to enable usage display.
        int usage = 0;

        // Pointers to output (capture) path names and files, and the count of blocks received, per channel.
        // These will be NULL if the corresponding channel is not to be captured.
        char* outpath[G7SERVER_MAXIMUM_CHANNELS];
        FILE* outfile[G7SERVER_MAXIMUM_CHANNELS];
        int multi_file_capture[G7SERVER_MAXIMUM_CHANNELS];
        unsigned int blocks_received;
        unsigned int first_block_tick;

        // Pointers to input (playback) path names and the data sourced from them.
        // These will be NULL if the corresponding channel is to be sourced with an internally-generated pattern rather than from a file.
        char* inpath[G7SERVER_MAXIMUM_CHANNELS];
        sample_t* inpattern[G7SERVER_MAXIMUM_CHANNELS];

        // General work variables.
        int arg;
        unsigned int level;
        unsigned int channel;
        unsigned int sample;
        int ret;
        int pass = 0;

        // Timeouts (-1 if not set).
        int connect_timeout = -1;
        int progress_timeout = -1;

        // Scratch buffer for reading capture data.
        sample_t* scratch = NULL;

        // Used to assemble pointers for g7_stage_output.
        const sample_t* source[G7SERVER_MAXIMUM_CHANNELS];

        // Default is not to save output.
        memset(outpath, 0, sizeof(char*) * G7SERVER_MAXIMUM_CHANNELS);
        memset(outfile, 0, sizeof(FILE*) * G7SERVER_MAXIMUM_CHANNELS);
        memset(multi_file_capture, 0, sizeof(int) * G7SERVER_MAXIMUM_CHANNELS);
        blocks_received = 0;
        first_block_tick = 0;

        // Default is to generate output patterns internally.
        memset(inpath, 0, sizeof(char*) * G7SERVER_MAXIMUM_CHANNELS);
        memset(inpattern, 0, sizeof(sample_t*) * G7SERVER_MAXIMUM_CHANNELS);

        // Default is not to analyze anything.
        unsigned int analyze_channel = (unsigned int)-1;

        setup_follower();

        // Process arguments except the last (must always be a value after the argument).
        for (arg = 1; arg < argc; arg++) {
                int val;
                int remaining = argc - arg;

                // All arguments must start with / or -, and are usually followed by an integer.
                if ((*argv[arg] != '-') && (*argv[arg] != '/')) {
                        usage = 1;
                        break;
                }
                if (remaining > 1) {
                        val = strtol(argv[arg + 1], NULL, 0);
                } else {
                        val = 0;
                }

                // Arguments requiring one parameter
                if (remaining >= 2) {

                        // /K <rate>
                        if (strcasecmp(argv[arg] + 1, "K") == 0) {
                                if (val < 0) {
                                        printf("Sample rate cannot be negative.\n");
                                        exit(1);
                                }
                                sample_rate = val;
                                arg++;
                                continue;
                        }

                        // /Q <blocks>
                        if ((strcasecmp(argv[arg] + 1, "Q") == 0) || (strcasecmp(argv[arg] + 1, "QQ") == 0)) {
                                if (val < 0) {
                                        printf("Starting quiet block count cannot be negative.\n");
                                        exit(1);
                                }
                                front_quiet_blocks = val;
                                if (strcasecmp(argv[arg] + 1, "QQ") == 0) {
                                        sync_sample_check = 0;
                                }
                                arg++;
                                continue;
                        }

                        // /E <blocks>
                        if (strcasecmp(argv[arg] + 1, "E") == 0) {
                                if (val < 0) {
                                        printf("Ending quiet block count cannot be negative.\n");
                                        exit(1);
                                }
                                end_quiet_blocks = val;
                                arg++;
                                continue;
                        }

                        // /B <blocks>
                        else if (strcasecmp(argv[arg] + 1, "B") == 0) {
                                if (val < 1) {
                                        printf("Block count must be at least 1.\n");
                                        exit(1);
                                }
                                data_blocks = val;
                                arg++;
                                continue;
                        }
                       
                        // /F <frequency>
                        else if (strcasecmp(argv[arg] + 1, "F") == 0) {
                                if (val < 1) {
                                        printf("Frequency must be a positive nonzero integer value in Hz.\n");
                                        exit(1);
                                }
                                // Nyquist criterion.
                                if (val > (SAMPLES_PER_SECOND / 2)) {
                                        printf("%d Hz is too high for the %d Hz sample rate.\n", val, SAMPLES_PER_SECOND);
                                        exit(1);
                                }
                                frequency = val;
                                arg++;
                                continue;
                        }

                        // /M <amplitude>
                        else if (strcasecmp(argv[arg] + 1, "M") == 0) {
                                if ((remaining < 1) || ((amplitude = atof(argv[arg + 1])) < 0.0) || (amplitude > 1.0)) {
                                        printf("Amplitude must be between 0.0 and 1.0\n");
                                        exit(1);
                                }
                                arg++;
                                continue;
                        }

                        // /TD
                        else if (strcasecmp(argv[arg] + 1, "TD") == 0) {
                                if ((remaining < 1) || ((amplitude = atof(argv[arg + 1])) < -1.0) || (amplitude > 1.0)) {
                                        printf("Level must be between -1.0 and 1.0\n");
                                        exit(1);
                                }
                                test_level = 1;
                                arg++;
                                continue;
                        }

                        // /TS
                        else if (strcasecmp(argv[arg] + 1, "TS") == 0) {
                                if ((remaining < 1) || ((stop_ms = atoi(argv[arg + 1])) < 100)) {
                                        printf("millisecond count for ts must be at least 100\n");
                                        exit(1);
                                }
                                printf("stop_ms = %u\n", stop_ms);
                                arg++;
                                continue;
                        }

                        // /S <samples>
                        else if (strcasecmp(argv[arg] + 1, "S") == 0) {
                                if ((val < G7SERVER_MIN_BLOCK_SAMPLES) && (val > G7SERVER_MAX_BLOCK_SAMPLES)) {
                                        printf("Samples per block must be in the range %d..%d.\n", G7SERVER_MIN_BLOCK_SAMPLES, G7SERVER_MAX_BLOCK_SAMPLES);
                                        exit(1);
                                }
                                samples_per_block = val;
                                arg++;
                                continue;
                        }
                       
                        // /C<ch> <path>
                        else if ((argv[arg][1] == 'C') || (argv[arg][1] == 'c') || (argv[arg][1] == 'U') || (argv[arg][1] == 'u')) {
                                unsigned int channel = atoi(argv[arg] + 2);
                                char* path = argv[arg + 1];
                                if (channel >= G7SERVER_MAXIMUM_CHANNELS) {
                                        printf("Capture (/C or /U) channel must be in the range 0..%d.\n", (int)(G7SERVER_MAXIMUM_CHANNELS - 1));
                                        exit(1);
                                }
                                if ((strlen(path) == 0) || (path[0] == '-') || (path[0] == '/')) {
                                        printf("\"%s\" doesn't look like a valid filename.\n", path);
                                        exit(1);
                                }
                                outpath[channel] = argv[arg + 1];
                                multi_file_capture[channel] = ((argv[arg][1] == 'U') || (argv[arg][1] == 'u'));
                                arg++;
                                continue;
                        }

                        // /L <loopback level>
                        else if ((strcasecmp(argv[arg] + 1, "L") == 0) && (arg < (argc - 1))) {
                                loopback = loopback_by_tag(argv[arg + 1])->val;
                                arg++;
                                continue;
                        }

                        // /A <address>
                        else if ((strcasecmp(argv[arg] + 1, "A") == 0) && (arg < (argc - 1))) {
                                address = argv[arg + 1];
                                arg++;
                                continue;
                        }

                        // /R<ch> <path>
                        else if ((argv[arg][1] == 'R') || (argv[arg][1] == 'r')) {
                                unsigned int channel = atoi(argv[arg] + 2);
                                char* path = argv[arg + 1];
                                if (channel >= G7SERVER_MAXIMUM_CHANNELS) {
                                        printf("Playback (/R) channel must be in the range 0..%d.\n", (int)(G7SERVER_MAXIMUM_CHANNELS - 1));
                                        exit(1);
                                }
                                if ((strlen(path) == 0) || (path[0] == '-') || (path[0] == '/')) {
                                        printf("\"%s\" doesn't look like a valid filename.\n", path);
                                        exit(1);
                                }
                                inpath[channel] = argv[arg + 1];
                                arg++;
                                continue;
                        }

                        // /D <int>
                        else if ((strcasecmp(argv[arg] + 1, "D") == 0) && (arg < (argc - 1))) {
                                g7client_debug_mask = (unsigned int)val;
                                arg++;
                                continue;
                        }

                        // /G <path>
                        else if ((strcasecmp(argv[arg] + 1, "G") == 0) && (arg < (argc - 1))) {
                                g7client_log_path = argv[arg + 1];
                                arg++;
                                continue;
                        }

                        // /J <blocks>
                        else if (strcasecmp(argv[arg] + 1, "J") == 0) {
                                junk_blocks = val;
                                arg++;
                                continue;
                        }

                        // /Y
                        else if ((strcasecmp(argv[arg] + 1, "Y") == 0) && (arg < (argc - 1))) {
                                if ((strlen(argv[arg + 1]) == 0) || (argv[arg + 1][0] == '-') || (argv[arg + 1][0] == '/') || parse_hardware_settings(argv[arg + 1], &new_state_data, &new_state_mask)) {
                                        printf("\"%s\" isn't a valid string of hardware setting(s).\n", argv[arg + 1]);
                                        exit(1);
                                }
                                arg++;
                                continue;
                        }

                        // /O
                        else if ((strcasecmp(argv[arg] + 1, "O") == 0) && (arg < (argc - 1))) {
                                oc_cmd = argv[arg + 1];
                                arg++;
                                continue;
                        }
                }
       
                // Arguments requiring no parameters.
                if (remaining >= 1) {

                        // /H
                        if (strcasecmp(argv[arg] + 1, "H") == 0) {
                                usage = 1;
                                break;
                        }

                        // /TF
                        else if (strcasecmp(argv[arg] + 1, "TF") == 0) {
                                test_pattern = 1;
                                continue;
                        }

                        // /TU
                        else if (strcasecmp(argv[arg] + 1, "TU") == 0) {
                                synchronize = 0;
                                continue;
                        }

                        // /I
                        else if (strcasecmp(argv[arg] + 1, "I") == 0) {
                                printf("Setting infinite timeouts for debugging.\n");
                                progress_timeout = 0;
                                connect_timeout = 0;
                                continue;
                        }

                        // /W
                        else if (strcasecmp(argv[arg] + 1, "W") == 0) {
                                printf("Requesting ramp-up and ramp-down blocks.\n");
                                ramp_blocks = 1;
                                continue;
                        }

                        // /Z
                        else if (strcasecmp(argv[arg] + 1, "Z") == 0) {
                                silence = 1;
                                continue;
                        }

                        // /P
                        else if (strcasecmp(argv[arg] + 1, "P") == 0) {
                                promiscuous = 1;
                                continue;
                        }

                        // /N<ch>
                        else if ((argv[arg][1] == 'N') || (argv[arg][1] == 'n')) {
                                unsigned int channel = atoi(argv[arg] + 2);
                                if (channel >= G7SERVER_MAXIMUM_CHANNELS) {
                                        printf("Capture (/C) channel must be in the range 0..%d.\n", (int)(G7SERVER_MAXIMUM_CHANNELS - 1));
                                        exit(1);
                                }
                                analyze_channel = (unsigned int)channel;
                                continue;
                        }

                        // /X
                        else if (strcasecmp(argv[arg] + 1, "X") == 0) {
                                extended = 1;
                                continue;
                        }

                        // /XL
                        else if (strcasecmp(argv[arg] + 1, "XL") == 0) {
                                extended = 2;
                                continue;
                        }

                        // /V
                        else if (strcasecmp(argv[arg] + 1, "V") == 0) {
                                printf("g7play " G7PLAY_VERSION " built " __DATE__ " " __TIME__ "\n");
                                printf("Supports Otostat-B and Otostat-C3 compatible hardware (including Rev E)\n");
                                exit(0);
                        }
                }

                // Unrecognized arguments.
                usage = 1;
                break;
        }

        if (usage) {
                printf("Usage: g7play [options]\n");
                printf("    Options can be upper or lower case, prefixed with / or -\n");
                printf("    -h or /? prints usage information\n");
                printf("    -k <bps> specifies the codec sample rate (24000 or 48000)\n");
                printf("    -b <blocks> specifies the number of blocks to send (default = %d)\n", DEFAULT_DATA_BLOCKS);
                printf("    -q <blocks> specifies the number of blocks of quiet to send before data blocks (default = %d)\n", DEFAULT_FRONT_QUIET_BLOCKS);
                printf("    -qq <blocks> is the same as -q, but suppresses minimum-sync-samples check\n");
                printf("    -e <blocks> specifies the number of blocks of quiet to send after data blocks (default = %d)\n", DEFAULT_END_QUIET_BLOCKS);
                printf("    -s <samples> specifies the number of samples per block (%d..%d, default = %d)\n", G7SERVER_MIN_BLOCK_SAMPLES, G7SERVER_MAX_BLOCK_SAMPLES, DEFAULT_SAMPLES_PER_BLOCK);
                printf("    -f <hz> specifies the approximate sine wave frequency (default = %d Hz)\n", DEFAULT_FREQUENCY);
                printf("    -m <amp> specifies the sine wave amplitude (0.0 to 1.0) (default = 1.0)\n");
                printf("    -l <keyword> selects loopback (default = %s), choices are:\n", loopback_by_value(DEFAULT_LOOPBACK)->tag);
                for (level = 0; level < num_loopback_levels; level++) {
                        printf("        %s (%s)\n", loopback_levels[level].tag, loopback_levels[level].desc);
                }
                printf("    -a <ipaddr:port> specifies IP address and port (default = %s)\n", G7SERVER_DEFAULT_ADDRESS);
                printf("    -cX saves captured audio from channel X to the specified file\n");
                printf("    -uX saves each block N of captured audio from channel X to a separate file named prefix.N\n");
                printf("    -rX reads playback audio for channel X from specified file\n");
                printf("    -tf uses framing pattern instead of sine wave\n");
                printf("    -td <level> uses DC level instead of sine wave, <level> is -1.0..1.0\n");
                printf("    -ts <ms> prematurely stop transfer <ms> milliseconds after first block received\n");
                printf("    -tu (unsynchronized) disables synchronization of capture and playback blocks\n");
                printf("    -z (zero) uses silence instead of sine wave\n");
                printf("    -w requests ramp-up and ramp-down blocks\n");
                printf("    -i specifies infinite timeouts for server communication, for debugging\n");
                printf("    -nX analyze input channel X (only one channel may be analyzed)\n");
                printf("    -g <path> specifies log path (defaults to stdout if not specified)\n");
                printf("    -d <number> specifies g7client debug mask (defaults to 0 if not specified)\n");
                printf("    -p promiscuous mode: capture response to all output, even quiet and ramp blocks\n");
                printf("    -x send single extended sequence; requires -R; ignores -B -F -M -T -Z -W -N\n");
                printf("    -xl like -x, with the first block of the sequence flagged as the start of a loop.\n");
                printf("    -j <blocks> specifies the number of post-ramp-up blocks to discard (junk) (default = %d)\n", DEFAULT_JUNK_BLOCKS);
                printf("       (if no ramp-up block is used, the first N blocks will be discarded)\n");
                printf("       (no blocks will be discarded when -P or -X is s)\n");
                printf("    -v displays version information\n");
                printf("    -y <keyword(s)> modifies Otostat-C3 audio setup prior to running audio\n");
                printf("       ppon, ppoff (probe power on/off), mbon, mboff (mic bias on/off), mpon, mpoff (mic preamp on/off)\n");
                printf("       moff, mlo, mhi (mic amp disabled, enabled w/low gain, enabled w/high gain)\n");
                printf("       auxoff, auxe, auxa, auxb (aux = off, aux = external input, aux = monitor a, aux = monitor b)\n");
                printf("       aoff, alo, ahi (receiver A driver off, enabled w/attenuation, enabled w/o attenuation)\n");
                printf("       boff, blo, bhi (receiver B driver off, enabled w/attenuation, enabled w/o attenuation)\n");
                printf("    -o <keyword> sets Otostat-C3 master state and exits (one keyword only)\n");
                printf("       Keywords are: off; standby; run (at rate specified by -k); query\n");
                exit(0);
        }

        // Generate a sine wave in a buffer.
        samples_per_second = (double)sample_rate;
        seconds_per_block = samples_per_block / samples_per_second;
        periods_per_second = (double)frequency;
        periods_per_block = periods_per_second * seconds_per_block;
        // Round periods per block to an integer, and back-compute the actual frequency.
        periods_per_block = (double)((int)(periods_per_block + 0.5));
        sine = (sample_t*)malloc(samples_per_block * sizeof(sample_t));
        for (sample = 0; sample < samples_per_block; sample++) {

                // Compute angular phase in radians.
                double phase = ((double)sample / (double)samples_per_block) * periods_per_block * M_PI * 2.0;

                // Scale the value to a signed 24-bit sample.
                // NOTE: assumes sizeof(sample_t) == 4.
                double normalized = amplitude * sin(phase) * (double)((1 << 23) - 1);

                // Left-justify and store the value.
                sine[sample] = ((sample_t)normalized) << 8;
        }

        // Generate a test pattern (or silence) in a buffer.
        test = (sample_t*)malloc(samples_per_block * sizeof(sample_t));
        if (silence) {
                memset(test, 0, samples_per_block * sizeof(sample_t));
        } else {
                for (sample = 0; sample < samples_per_block; sample++) {
                        if (test_level) {
                                double normalized = amplitude * (double)((1 << 23) - 1);
                                test[sample] = ((sample_t)normalized) << 8;
                        } else {
                                // 7FFF in the top of the word, and the sample number in the next 16 bits.
                                test[sample] = 0x7FFF0000 | sample;
                        }
                }
        }

        // A temporary data buffer.
        scratch = (sample_t*)malloc(BLOCKS_PER_READ * G7SERVER_MAXIMUM_CHANNELS * samples_per_block * sizeof(sample_t));

        // The handle to the G7 client instance.
        g7handle_t h = NULL;

        // Set if we actually connected, and started an audio transfer.
        int connected = 0;
        int started = 0;

        // Stop and uninitialize on exit from this loop.
        for (;;) {
                int fail;
                int work;

                // Initialize.
                h = g7_init(g7client_debug_mask, g7client_log_path);
                if (h == NULL) {
                        printf("g7_init: %s\n", g7_errtext(h));
                        break;
                }

                // Set block size.

                // Set timeouts.
                if (connect_timeout >= 0) {
                        ret = g7_control(h, G7_FLAG_CTRL_CONNECT_TIMEOUT, connect_timeout);
                        if (ret != G7_OK) {
                                printf("g7_control (connect timeout): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                break;
                        }
                }
                if (progress_timeout >= 0) {
                        ret = g7_control(h, G7_FLAG_CTRL_PROGRESS_TIMEOUT, progress_timeout);
                        if (ret != G7_OK) {
                                printf("g7_control (progress timeout): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                break;
                        }
                }

                // Set block size.
                ret = g7_control(h, G7_FLAG_CTRL_SAMPLES_PER_BLOCK, samples_per_block);
                if (ret != G7_OK) {
                        printf("g7_control (samples per block): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }

                // Connect to the server.
                ret = g7_connect(h, address);
                if (ret != G7_OK) {
                        printf("g7_connect: %s: %s.\n", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }
                connected = 1;

                // Assume Rev B hardware.
                int rev_c3 = 0;

                // Get the current state of the audio hardware.
                unsigned int old_state_mask = 0;
                unsigned int old_state_data = 0;
                ret = g7_query(h, G7_FLAG_QUERY_SETUP_C3, &old_state_data, &old_state_mask, NULL, NULL);
                if (ret == G7_OK) {
                        rev_c3 = 1;
                        printf("Rev C3+ hardware detected.\n");
                        printf("Current hardware settings: ");
                        interpret_hardware_settings(old_state_data, old_state_mask);
                        printf("\n");
                } else if (ret == G7_ERR_SERVER) {
                        // If g7server refused the command, it must be Rev B.
                        printf("Rev C3+ features not supported; assuming Rev B hardware.\n");
                } else {
                        printf("g7_query (setup): %s: %s.\n", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }

                // If not Rev C3, reject unsupported command.
                if (rev_c3 == 0) {
                        if (new_state_mask != 0) {
                                printf("-y option requires Rev C3 hardware.\n");
                                break;
                        }
                        if (oc_cmd != NULL) {
                                printf("-o option requires Rev C3 hardware.\n");
                                break;
                        }
                }

                // If Rev C3, set up audio path.
                 else {
                        // If the user specified changes to the hardware state, apply them.
                        new_state_data = (old_state_data & old_state_mask & ~new_state_mask) | (new_state_data & old_state_mask & new_state_mask);

                        // Default to the same rate specified with the -k option.
                        // Note that for Rev C3, this overrides the current state; the user _must_ specify a sample rate every time unless they want the default (48 kHz).
                        int fail = 0;
                        new_state_data &= ~G7_SETUP_C3_MASTER_MASK;
                        switch (sample_rate) {
                        case 16000:
                                new_state_data |= G7_SETUP_C3_MASTER_16KHZ;
                                break;
                        case 24000:
                                new_state_data |= G7_SETUP_C3_MASTER_24KHZ;
                                break;
                        case 32000:
                                new_state_data |= G7_SETUP_C3_MASTER_32KHZ;
                                break;
                        case 48000:
                                new_state_data |= G7_SETUP_C3_MASTER_48KHZ;
                                break;
                        default:
                                printf("Sample rate must be 16000, 24000, 32000, or 48000.\n");
                                fail = 1;
                        }
                        if (fail) {
                                break;
                        }

                        // Perform do-it-and-exit operations.
                        if (oc_cmd != NULL) {
                                // Power state?
                                if (0 == strcasecmp(oc_cmd, "off")) {
                                        new_state_data &= ~G7_SETUP_C3_MASTER_MASK;
                                        new_state_data |= G7_SETUP_C3_MASTER_OFF;
                                }
                                else if (0 == strcasecmp(oc_cmd, "standby")) {
                                        new_state_data &= ~G7_SETUP_C3_MASTER_MASK;
                                        new_state_data |= G7_SETUP_C3_MASTER_STANDBY;
                                }
                                else if (0 == strcasecmp(oc_cmd, "run")) {
                                        // This will be selected by default.
                                }
                                else if (0 == strcasecmp(oc_cmd, "query")) {
                                        // We already showed the settings, so just exit.
                                        // (If the debug code above goes away, save the state and print it here).
                                        break;
                                }
                                else {
                                        printf("Invalid -o option specified; choices are off, standby, run, query\n");
                                        break;
                                }
                        }
                        printf("New hardware settings:     ");
                        interpret_hardware_settings(new_state_data, old_state_mask);
                        printf("\n");

                        // Apply the hardware settings.
                        ret = g7_control(h, G7_FLAG_CTRL_SETUP_C3, new_state_data);
                        if (ret != G7_OK) {
                                printf("g7_control (setup): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                break;
                        }

                        // If a do-it-and-exit operation was requested, we are done.
                        if (oc_cmd != NULL) {
                                break;
                        }
                }

                // Set the sample rate.
                ret = g7_control(h, G7_FLAG_CTRL_RATE, sample_rate);
                if (ret != G7_OK) {
                        g7_connect(h, NULL);
                        printf("g7_control (rate): %s: %s", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }

                // Get the channel counts.
                unsigned int server_playback_channels = 0;
                unsigned int server_capture_channels = 0;
                ret = g7_query(h, G7_FLAG_QUERY_CHANNELS, &server_playback_channels, &server_capture_channels, NULL, NULL);
                if (ret != G7_OK) {
                        printf("g7_query (channels): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }

                // Ensure that we have at least SYNC_SAMPLES worth of quiet and/or ramp-up blocks.
                if (extended) {
                        if (ramp_blocks) {
                                printf("Warning: ramp-up and ramp-down not supported in extended mode; ignoring\n");
                                ramp_blocks = 0;
                        }
                }
                if (sync_sample_check) {
                        unsigned int header_samples = samples_per_block * (front_quiet_blocks + ramp_blocks) * server_playback_channels;
                        if (header_samples < SYNC_SAMPLES) {
                                unsigned int extra_quiet_blocks = ((SYNC_SAMPLES - header_samples) / (samples_per_block * server_playback_channels)) + 1;
                                printf("Adding %d quiet block%s to meet synchronization requirements\n", extra_quiet_blocks, (extra_quiet_blocks > 1) ? "s" : "");
                                front_quiet_blocks += extra_quiet_blocks;
                        }
                }

                // Open capture files for single-file capture as requested.
                for (channel = 0; channel < server_capture_channels; channel++) {
                        if ((outpath[channel] != NULL) && (multi_file_capture[channel] == 0)) {
                                outfile[channel] = fopen(outpath[channel], "wb");
                                if (outfile[channel] == NULL) {
                                        printf("Could not open %s for writing: %s\n", outpath[channel], strerror(errno));
                                        break;
                                }
                                printf("Capturing channel %d to %s.\n", channel, outpath[channel]);
                        }
                }

                // Read playback files as requested.
                for (channel = 0; channel < server_playback_channels; channel++) {

                        if (inpath[channel] != NULL) {
                                FILE* infile = fopen(inpath[channel], "rb");
                                if (infile == NULL) {
                                        printf("Could not open %s for reading: %s\n", inpath[channel], strerror(errno));
                                        break;
                                }

                                // If playing an extended sequence, read the entire file.
                                if (extended) {
                                        int blocks;
                                        struct stat stats;
                                       
                                        // Get the file size.
                                        if (fstat(fileno(infile), &stats) != 0) {
                                                printf("Could not get file information on %s: %s\n", inpath[channel], strerror(errno));
                                                exit(1);
                                        }
                                        blocks = (stats.st_size / (sizeof(sample_t) * samples_per_block));
                                        if ((blocks * sizeof(sample_t) * samples_per_block) != stats.st_size) {
                                                int remnant = stats.st_size - (blocks * sizeof(sample_t) * samples_per_block);
                                                printf("%s contains %d full blocks and a remnant of %d bytes\n", inpath[channel], blocks, remnant);
                                                blocks++;
                                        } else {
                                                printf("%s contains %d full blocks\n", inpath[channel], blocks);
                                        }
                                        if (blocks == 0) {
                                                printf("%s is empty\n", inpath[channel]);
                                                exit(1);
                                        }

                                        // Adjust all lengths to the length of the first playback file encountered.
                                        if (extended_blocks == 0) {
                                                printf ("The length of the output sequence will be set to %d blocks\n", blocks);
                                                extended_blocks = blocks;
                                        } else if (blocks > extended_blocks) {
                                                printf("%s is longer than first playback file; truncating to %d blocks\n", inpath[channel], extended_blocks);
                                                blocks = extended_blocks;
                                        } else if (blocks < extended_blocks) {
                                                printf("%s is shorter than first playback file; will be silence-padded to %d blocks\n", inpath[channel], extended_blocks);
                                                blocks = extended_blocks;
                                        }

                                        // Read the file.
                                        inpattern[channel] = new sample_t[blocks * samples_per_block];
                                        memset(inpattern[channel], 0, blocks * samples_per_block * sizeof(sample_t));

                                        // CEA 2012/04/19: mingw can't seem to read more than 32K bytes with fread() or read();
                                        // the call fails with EINVAL.  So we have to break the call up into multiple reads.
                                        ssize_t bytes_left = stats.st_size;

                                        char* dest = (char*)(inpattern[channel]);

                                        while (bytes_left > 0) {
                                                ssize_t pass_bytes = ((unsigned int)bytes_left > (samples_per_block * sizeof(sample_t))) ?
                                                        samples_per_block * sizeof(sample_t) : bytes_left;

                                                //read from infile and write to inpattern a 'char' at a time, even though the infile is binary
                                                int r = fread(dest, 1, pass_bytes, infile);

                                                if (r < 0) {
                                                        printf("Error reading %s: %s\n", inpath[channel], strerror(errno));
                                                } else if (r != pass_bytes) {
                                                        printf("Short read from %s (read %d bytes, expected %d)\n", inpath[channel], r, (int)pass_bytes);
                                                        exit(1);
                                                }

                                                bytes_left -= pass_bytes;
                                                dest += pass_bytes;
                                        }
                                }

                                // For non-extended sequences, read a single block.
                                else {

                                        // Allocate buffer and read samples.
                                        inpattern[channel] = new sample_t[samples_per_block];
                                        int r = fread(inpattern[channel], sizeof(sample_t), samples_per_block, infile);
                                        if (r < 0) {
                                                printf("Error reading %s: %s\n", inpath[channel], strerror(errno));
                                                exit(1);
                                        } else if (r != (int)samples_per_block) {
                                                printf("%s is the wrong length (expected %d samples, got %d)\n", inpath[channel], samples_per_block, r);
                                                exit(1);
                                        }
                                }

                                fclose(infile);
                        }
                }

                ret = g7_control(h, G7_FLAG_CTRL_LOOPBACK, loopback);
                if (ret != G7_OK) {
                        printf("g7_control (loopback): %s: %s.\n", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }

                // Allocate a unique tag for each block we append to the sequence.
                int tag = 1;

                // Set data.
                fail = 0;
                for (channel = 0; channel < server_playback_channels; channel++) {
                        sample_t* data = NULL;
                        if (inpattern[channel] != NULL) {
                                printf("Channel %d stimulus data is external input.\n", channel);
                                data = inpattern[channel]; // data now holds the (pointer to) the inpattern for the channel
                        } else if (test_pattern) {
                                printf("Channel %d stimulus data is non-audio framing pattern.\n", channel);
                                // test buffer will have been filled with test pattern
                                data = test;
                        } else if (test_level) {
                                double normalized = amplitude * (double)((1 << 23) - 1);
                                sample_t sample = ((sample_t)normalized) << 8;
                                printf("Channel %d stimulus data is DC level (0x%08X, approximately %.2fV).\n", channel, sample, 2.5f + (amplitude * 2.5f));
                                // test buffer will have been filled with test pattern
                                data = test;
                        } else if (silence) {
                                printf("Channel %d stimulus data is silence\n", channel);
                                data = NULL;
                        } else if (extended) {
                                // Only an external pattern or silence is allowed in extended-sequence mode.
                                printf("In extended mode, channel stimulus must be external data or silence; channel %d is not\n", channel);
                                exit(1);
                        } else {
                                printf("Channel %d stimulus data is a %.2f Hz sine wave.\n", channel, periods_per_block / seconds_per_block);
                                data = sine;
                        }
                        source[channel] = data; //now source[channel] holds the data to be played
                }

                // Reset the output sequence.
                ret = g7_stage_output(h, G7_FLAG_SO_RESET, 0, NULL, 0, 0);
                if (ret != G7_OK) {
                        printf("g7_stage_output (reset): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }

                // Add quiet blocks.
                if (front_quiet_blocks != 0) {
                        ret = g7_stage_output(h, G7_FLAG_SO_APPEND_DATA | (promiscuous ? 0 : G7_FLAG_SO_NO_CAPTURE), server_playback_channels, NULL, front_quiet_blocks, tag++);
                }

                // Add ramp-up block.
                if (ramp_blocks) {
                        ret = g7_stage_output(h, G7_FLAG_SO_APPEND_RAMP_UP | (promiscuous ? 0 : G7_FLAG_SO_NO_CAPTURE), server_playback_channels, source, 1, tag++);
                        if (ret != G7_OK) {
                                printf("g7_stage_output (ramp-up): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                break;
                        }
                }

                if (extended) {
                        // Add extended data.
                        for (int block = 0; block < extended_blocks; block++) {

                                // source will point to the beginning of the data for each channel, or NULL for silence.
                                // block_source will be set to point to the beginning of each particular block.
                                const sample_t* block_source[G7SERVER_MAXIMUM_CHANNELS];
                                for (unsigned int channel = 0; channel < server_playback_channels; channel++) {
                                        if (source[channel] == NULL) {
                                                // silence
                                                block_source[channel] = NULL;
                                        } else {
                                                // actual data
                                                //pointer to the data starts at the beginning and moves samples_per_block through the data
                                                block_source[channel] = source[channel] + (block * samples_per_block);
                                        }
                                }

                                // If -xl was specified, mark the firsthis is the first ft block of the sequence as start-of-loop.
                                unsigned int flags = G7_FLAG_SO_APPEND_DATA;
                                if ((extended == 2) && (block == 0)) {
                                        flags |= G7_FLAG_SO_START_OF_LOOP;
                                }

                                // Append this block worth of samples to the staging sequence, to be played once.
                                // Don't increment the tag counter since it forces us into block-at-a-time read mode.
                                ret = g7_stage_output(h, flags, server_playback_channels, block_source, 1, tag);
                                if (ret != G7_OK) {
                                        printf("g7_stage_output (data block %d/%d): %s: %s\n", block + 1, extended_blocks, g7_returntext(h, ret), g7_errtext(h));
                                        fail = 1;
                                        break;
                                }
                        }
                        if (fail) {
                                break;
                        }
       
                        // Increment the tag number before we output the quiet blocks.
                        tag++;
                }

                else {
                        // Add the blocks to be junked.
                        if (junk_blocks > 0) {
                                ret = g7_stage_output(h, G7_FLAG_SO_APPEND_DATA | (promiscuous ? 0 : G7_FLAG_SO_NO_CAPTURE), server_playback_channels, source, junk_blocks, tag++);
                                if (ret != G7_OK) {
                                        printf("g7_stage_output (junk): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                        break;
                                }
                        }

                        // Add one block worth of samples, to be repeated data_blocks times.
                        ret = g7_stage_output(h, G7_FLAG_SO_APPEND_DATA, server_playback_channels, source, data_blocks, tag++);
                        if (ret != G7_OK) {
                                printf("g7_stage_output (data): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                break;
                        }
                }

                // Add ramp-down block.
                if (ramp_blocks) {
                        ret = g7_stage_output(h, G7_FLAG_SO_APPEND_RAMP_DOWN | (promiscuous ? 0 : G7_FLAG_SO_NO_CAPTURE), server_playback_channels, source, 1, tag++);
                        if (ret != G7_OK) {
                                printf("g7_stage_output (ramp-down): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                break;
                        }
                }

                // Add more quiet blocks.
                if (end_quiet_blocks) {
                        ret = g7_stage_output(h, G7_FLAG_SO_APPEND_DATA | (promiscuous ? 0 : G7_FLAG_SO_NO_CAPTURE), server_playback_channels, NULL, end_quiet_blocks, tag++);
                        if (ret != G7_OK) {
                                printf("g7_stage_output (end-quiet): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                break;
                        }
                }

                // Describe the server and the loopback mode.
                printf("Server at %s.\n", address);
                for (level = 0; level < num_loopback_levels; level++) {
                        if (loopback == loopback_levels[level].val) {
                                printf("%s.\n", loopback_levels[level].desc);
                        }
                }

                // Describe the data sequence.
                printf("Data sequence (all blocks have %d samples):\n", samples_per_block);
                const char* pm = (promiscuous ? " (*)" : "");
                if (front_quiet_blocks != 0) {
                        printf("  %d preceding quiet block%s%s\n", front_quiet_blocks, (front_quiet_blocks != 1) ? "s" : "", pm);
                }
                if (extended) {
                        printf("  %d block%s of %s extended audio sequence (*)\n", extended_blocks,
                                (extended_blocks != 1) ? "s" : "",
                                (extended == 2) ? "a looped" : "an");
                } else {
                        if (ramp_blocks != 0) {
                                printf("  1 ramp-up block%s\n", pm);
                        }
                        if (junk_blocks != 0) {
                                printf("  %d junk block%s%s\n", junk_blocks, (junk_blocks != 1) ? "s" : "", pm);
                        }
                        if (data_blocks != 0) {
                                printf("  %d data block%s (*)\n", data_blocks, (data_blocks != 1) ? "s" : "");
                        }
                        if (ramp_blocks != 0) {
                                printf("  1 ramp-down block%s\n", pm);
                        }
                }
                if (end_quiet_blocks != 0) {
                        printf("  %d following quiet block%s%s\n", end_quiet_blocks, (end_quiet_blocks != 1) ? "s" : "", pm);
                }
                printf("Response will be captured during portions of sequence marked with (*)\n");
                if (synchronize == 0) {
                        printf("WARNING: Synchronization suppressed; capture blocks will be unaligned with playback blocks\n");
                }
       
                // Start playback and capture.
                ret = g7_control_output(h, synchronize ? G7_FLAG_CO_START : G7_FLAG_CO_START_UNSYNCHRONIZED);
                if (ret != G7_OK) {
                        printf("g7_control_output (start): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }
                printf("In progress.\n");
                started = 1;

                // Wait for playback to stop, and all of our capture data arrive.
                // work will be cleared when we find nothing to do for either channel.
                work = 1;
                while (work) {

                        // Assume there's nothing to do next time.
                        work = 0;

                        // Wait a little while.
#if defined(_MSC_VER)
                        ::Sleep(50);    // millseconds
#else
                        usleep(50000);  // microseconds
#endif

                        for (;;) {
                                unsigned int blocks;
                                unsigned int bitmap;
                                const sample_t* src;

                                // Check again next time.
                                work = 1;

                                // Fill with a distinguishable pattern (0xCD) so we make sure we're getting real data.
                                memset(scratch, 0xCD, server_capture_channels * samples_per_block * sizeof(sample_t));

                                // Get capture data.
                                blocks = BLOCKS_PER_READ;
                                bitmap = (server_capture_channels > 31) ? 0xFFFFFFFFUL : ((1U << server_capture_channels) - 1);
                                unsigned int tag, sequence;
                                int err = g7_read_input(h, bitmap, scratch, &blocks, 0, &tag, &sequence);
                                if (err != G7_OK) {
                                        printf("g7_read_input: %s: %s\n", g7_returntext(h, err), g7_errtext(h));

                                        // Bail out of the work loop.
                                        work = 0;
                                        break;
                                }

                                // If we got no blocks, bail out and wait a while.
                                if (blocks == 0) {
                                        break;
                                }
                                printf("got %d blocks from g7_read_input (tag 0x%08x, sequence %d..%d)\n", blocks, tag, sequence, sequence + blocks - 1);

                                // Process blocks.
                                src = scratch;
                                while (blocks != 0) {

                                        // Keep track of the time the first block was received.
                                        if (blocks_received == 0) {
                                                // Add 1 to the tick value so it can't end up being 0.
                                                first_block_tick = tick() + 1;
                                        }

                                        // Keep track of the total number of blocks received.
                                        blocks_received++;

                                        // Process each channel.
                                        for (channel = 0; channel < server_capture_channels; channel++) {

                                                // If doing multi-file capture, need to open a new file for each block.
                                                if ((multi_file_capture[channel] != 0) && (outpath[channel] != NULL)) {
                                                        char single_outpath[MAXPATHLEN];
                                                        char block_number_text[20];
                                                        strcpy(single_outpath, outpath[channel]);
                                                        strcat(single_outpath, ".");
                                                        sprintf(block_number_text, "%d", blocks_received);
                                                        strcat(single_outpath, block_number_text);
                                                        outfile[channel] = fopen(single_outpath, "wb");
                                                        if (outfile[channel] == NULL) {
                                                                printf("Could not open %s for writing: %s\n", outpath[channel], strerror(errno));
                                                                break;
                                                        }
                                                }

                                                // Write to file if requested.
                                                if (outfile[channel] != NULL) {
                                                        if (fwrite(src, sizeof(sample_t), samples_per_block, outfile[channel]) != samples_per_block) {
                                                                printf("Error writing capture data from channel %d to file: %s\n", channel, strerror(errno));

                                                                // Bail out of the work loop.
                                                                work = 0;
                                                                break;
                                                        } else {
                                                                printf("Wrote block %d for channel %d\n", blocks_received, channel);
                                                                //printf("  %08X: %08X %08X\n", src, *(unsigned int*)src, *(unsigned int*)(((char*)src) + 4));
                                                        }
                                                }

                                                // If doing multi-file capture, need to close the file after each block.
                                                if ((multi_file_capture[channel] != 0) && (outfile[channel] != NULL)) {
                                                        fclose(outfile[channel]);
                                                        outfile[channel] = NULL;
                                                }

                                                // Analyze the channel if requested.
                                                if (channel == analyze_channel) {
                                                        printf("Analyzing block for channel %d\n", channel);
                                                        analyze_block((int)samples_per_block, src);
                                                }

                                                // Advance to the next channel's data or to the next block.
                                                src += samples_per_block;
                                        }

                                        blocks--;
                                }
                        }

                        // If we are supposed to stop playback, check for timeout.
                        if ((stop_ms != 0) && (first_block_tick != 0) && ((tick() - (first_block_tick - 1)) >= stop_ms)) {
                                printf("Stopping playback after %u ms\n", tick() - (first_block_tick - 1));

                                ret = g7_control_output(h, G7_FLAG_CO_STOP | G7_FLAG_CO_DISCARD);
                                if (ret != G7_OK) {
                                        // We can't really do anything if this fails.
                                        printf("g7_control_output (stop): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                                }

                                // Bail out of the work loop.
                                work = 0;
                        }
                }

                break;
        }

        // Wait for playback to stop, if started.
        pass = 0;
        while (started) {
                unsigned int state = G7_QUERY_PLAYBACK_UNKNOWN;
                unsigned int tag = 0xFFFFFFFFU;
                ret = g7_query(h, G7_FLAG_QUERY_PLAYBACK, &state, &tag, NULL, NULL);
                if (ret != G7_OK) {
                        // We can't really do anything if this fails.
                        printf("g7_query (playback): %s: %s\n", g7_returntext(h, ret), g7_errtext(h));
                        break;
                }

                if (state == G7_QUERY_PLAYBACK_HALTED) {
                        printf("Codec has halted\n");
                        break;
                }

                if (pass == 0) {
                        printf("Waiting for codec to stop playing\n");
                }

#if defined(_MSC_VER)
                ::Sleep(10);    // milliseconds
#else
                usleep(10000);
#endif

                if (pass++ == 10) {
                        printf("Waited too long for codec to stop playing\n");
                        break;
                }
        }

        // Disconnect, if connected.
        if (connected) {
                g7_connect(h, NULL);
        }

        // Clean up.
        printf("Shutting down.\n");
        g7_shutdown(h);
        cleanup_follower();

        for (channel = 0; channel < G7SERVER_MAXIMUM_CHANNELS; channel++) {
                if (outfile[channel] != NULL) {
                        fclose(outfile[channel]);
                }
        }

        free(sine);
        free(test);
        free(scratch);
        printf("Done.\n");

        return 0;
}
