//(root)/Gen7/HandheldSourceCode/Overlay/recipes/g7common/g7shared.h - Rev 13706
//
// Gen7 Shared Utilities
// Copyright (c) 2010-2015 Mimosa Acoustics, Inc.

// Linux end-of-line markers, 8-column _tabs_ (not spaces), K&R C brace style.

// SOURCE CONTROL WARNING:  The master version of this file is located in the Gen7 repository at
//      Gen7/HandheldSourceCode/Overlay/recipes/g7common/g7shared.h
// Subversion updates under Windows may result in duplicate copies due to following symbolic links,
// but all changes should be made and committed to the master version at the location showmn above.

#pragma once

// A sample.
typedef int sample_t;

// Common to all configurations.
#define G7SERVER_MIN_BLOCK_SAMPLES              16
#define G7SERVER_MAX_BLOCK_SAMPLES              65536
#define G7SERVER_DEFAULT_ADDRESS                "10.77.65.1:1024"
#define G7SERVER_DEFAULT_PORT                   1024
#define G7SERVER_DEFAULT_CONNECT_TIMEOUT_MS     2000
#define G7SERVER_DEFAULT_PROGRESS_TIMEOUT_MS    2000
#define G7SERVER_MAXIMUM_CHANNELS               (8 * sizeof(unsigned int))
#define G7SERVER_BATTERY_FILENAME               "/var/otostat/results/battery.txt"

#define G7CPSPISERVER_DEFAULT_PORT              1001

// Current g7client/g7server protocol version.
// The server will reject connections with a mismatching major protocol number.
#define PROTOCOL_VERSION_MAJOR  1U
#define PROTOCOL_VERSION_MINOR  0U

// Error codes.
#define G7_OK                           0       // No error
#define G7_ERR_INIT                     -1      // Client not initialized, or required value not set
#define G7_ERR_STATE                    -2      // Audio peripheral is not in appropriate state
#define G7_ERR_NETWORK                  -3      // Network error connecting to audio peripheral
#define G7_ERR_ARGS                     -4      // Value out of range
#define G7_ERR_VERSION                  -5      // Version mismatch between client and peripheral
#define G7_ERR_END_OF_STREAM            -6      // No more data available on this channel
#define G7_ERR_SERVER                   -7      // Error reported by the audio peripheral
#define G7_ERR_NOTCONNECTED             -8      // Connection to server required for this command
#define G7_ERR_CONNECTED                -9      // Connection to server precludes this command
#define G7_ERR_SHUTDOWN                 -10     // Audio path has not been powered up

// Argument values for g7_ctrl(G7_FLAG_CTRL_LOOPBACK)
#define G7_LOOPBACK_NONE                0       // no loop; reports audio received from codec (requires hardware codec)
#define G7_LOOPBACK_I2S                 1       // hard loop on digital path to codec (requires hardware codec)
#define G7_LOOPBACK_SERVER              2       // device simulates codec (no hardware codec required)

// Argument bits for g7_ctrl(G7_FLAG_CTRL_SETUP) for Rev C3 compatible hardware (Rev C3 and Rev E).

// External 5V probe power
#define G7_SETUP_C3_PROBE_POWER_MASK    0x04000000
#define G7_SETUP_C3_PROBE_POWER_OFF     0x00000000      // 5V external probe power (for differential probes) off (Rev C3)
#define G7_SETUP_C3_PROBE_POWER_ON      0x04000000      // 5V external probe power (for differential probes) on (Rev C3)

// Single-ended microphone bias
#define G7_SETUP_C3_MIC_BIAS_MASK       0x02000000
#define G7_SETUP_C3_MIC_BIAS_OFF        0x00000000      // Microphone bias (for single-ended probes) off (Rev C3)
#define G7_SETUP_C3_MIC_BIAS_ON         0x02000000      // Microphone bias (for single-ended probes) on (Rev C3)

// Single-ended microphone preamplifier
#define G7_SETUP_C3_MIC_PREAMP_MASK     0x01000000
#define G7_SETUP_C3_MIC_PREAMP_OFF      0x00000000      // Microphone preamplifier (for single-ended probes) disabled (Rev C3)
#define G7_SETUP_C3_MIC_PREAMP_ON       0x01000000      // Microphone preamplifier (for single-ended probes) eanbled (Rev C3)

// Primary audio input state
#define G7_SETUP_C3_PRIMARY_INPUT_MASK  0x00F00000
#define G7_SETUP_C3_PRIMARY_INPUT_OFF   0x00000000      // Primary audio input disabled (Rev C3, Rev E EAB)
#define G7_SETUP_C3_PRIMARY_INPUT_LOW   0x00100000      // Primary audio input enabled in low-gain mode (Rev C3, Rev E EAB)
#define G7_SETUP_C3_PRIMARY_INPUT_HIGH  0x00200000      // Primary audio input enabled in high-gain mode (Rev C3, Rev E EAB)
// Reserve value 3 for a possible third gain level for implementation in Rev E.

// Extra primary audio input state values for Rev E hardware, for engineering use only.
// Note that setting to a state which aliases back to a C3_PRIMARY_INPUT_ state will report that value instead.
// depending on how the input paths are configured on the Rev E board and in g7server.  
#define G7_SETUP_C3_E_PRIMARY_INPUT_M   0x00400000      // Primary audio input enabled w/main feedback path (Rev E EAB)
#define G7_SETUP_C3_E_PRIMARY_INPUT_MA  0x00500000      // Primary audio input enabled w/main and "A" feedback paths (Rev E EAB)
#define G7_SETUP_C3_E_PRIMARY_INPUT_MB  0x00600000      // Primary audio input enabled w/main and "B" feedback paths (Rev E EAB)
#define G7_SETUP_C3_E_PRIMARY_INPUT_MAB 0x00700000      // Primary audio input enabled w/main, "A" and "B" feedback paths (Rev E EAB)

// Auxiliary audio input state
#define G7_SETUP_C3_AUX_INPUT_MASK      0x000F0000
#define G7_SETUP_C3_AUX_INPUT_OFF       0x00000000      // Auxiliary audio input disabled (Rev C3, Rev E EAB)
#define G7_SETUP_C3_AUX_INPUT_EXTERNAL  0x00010000      // Auxiliary audio amplifier enabled, from external input (Rev C3, Rev E EAB)
#define G7_SETUP_C3_AUX_INPUT_MON_A     0x00020000      // Auxiliary audio amplifier enabled, from receiver A monitor (Rev C3, Rev E EAB)
#define G7_SETUP_C3_AUX_INPUT_MON_B     0x00030000      // Auxiliary audio amplifier enabled, from receiver B monitor (Rev C3, Rev E EAB)

// Receiver A driver state
#define G7_SETUP_C3_OUT_A_MASK          0x0000F000
#define G7_SETUP_C3_OUT_A_OFF           0x00000000      // Receiver A disabled (Rev C3, Rev E EAB)
#define G7_SETUP_C3_OUT_A_ATTENUATED    0x00001000      // Receiver A enabled with output attenuation (Rev C3, Rev E EAB)
#define G7_SETUP_C3_OUT_A_UNATTENUATED  0x00002000      // Receiver A enabled without output attenuation (Rev C3, Rev E EAB)

// Receiver B driver state
#define G7_SETUP_C3_OUT_B_MASK          0x00000F00
#define G7_SETUP_C3_OUT_B_OFF           0x00000000      // Receiver B disabled (Rev C3, Rev E EAB)
#define G7_SETUP_C3_OUT_B_ATTENUATED    0x00000100      // Receiver B enabled with output attenuation (Rev C3, Rev E EAB)
#define G7_SETUP_C3_OUT_B_UNATTENUATED  0x00000200      // Receiver B enabled without output attenuation (Rev C3, Rev E EAB)

// Master state
#define G7_SETUP_C3_MASTER_MASK         0x0000000F
#define G7_SETUP_C3_MASTER_OFF          0x00000000      // Audio block disabled (all other settings ignored) (Rev C3, All Rev E)
#define G7_SETUP_C3_MASTER_STANDBY      0x00000001      // Audio block in standby (all other settings ignored) (Rev C3, All Rev E)
#define G7_SETUP_C3_MASTER_16KHZ        0x00000002      // Audio block running at 16 KHz (Rev C3, All Rev E)
#define G7_SETUP_C3_MASTER_24KHZ        0x00000003      // Audio block running at 24 KHz (Rev C3, All Rev E)
#define G7_SETUP_C3_MASTER_32KHZ        0x00000004      // Audio block running at 32 KHz (Rev C3, All Rev E)
#define G7_SETUP_C3_MASTER_48KHZ        0x00000005      // Audio block running at 48 KHz (Rev C3, All Rev E)

// Flags for g7_ calls.
// NOTE: Flags must be non-overlapping only within the context of their specific command.

// g7_control
#define G7_FLAG_CTRL_CONNECT_TIMEOUT    0x00000001
#define G7_FLAG_CTRL_PROGRESS_TIMEOUT   0x00000002
#define G7_FLAG_CTRL_RATE               0x00000004
#define G7_FLAG_CTRL_LOOPBACK           0x00000008
#define G7_FLAG_CTRL_SAMPLES_PER_BLOCK  0x00000010
#define G7_FLAG_CTRL_CLIENT_DEBUG_MASK  0x00000020
#define G7_FLAG_CTRL_SETUP_C3           0x00000100      // Otostat-C only

// g7_stage_output
#define G7_FLAG_SO_APPEND_DATA          0x00000100
#define G7_FLAG_SO_APPEND_RAMP_UP       0x00000200
#define G7_FLAG_SO_APPEND_RAMP_DOWN     0x00000400
#define G7_FLAG_SO_RESET                0x00000800
#define G7_FLAG_SO_NO_CAPTURE           0x00001000
#define G7_FLAG_SO_HALT                 0x00002000      // see note
#define G7_FLAG_SO_TRANSITION           0x00004000      // see note
#define G7_FLAG_SO_START_OF_LOOP        0x00008000      // see note
// G7_FLAG_SO_HALT is used internal to the server as a way to flag fake blocks which are being sourced as
// a result of the output sequence being exhausted.  It is not possible for the user to specify the flag.
// G7_FLAG_SO_TRANSITION is used internal to the server to mark blocks during a transition in the codec
// loopback state.  It is not possible for the user to specify the flag.
// G7_FLAG_SO_START_OF_LOOP indicates the start of a sequence of blocks (a loop).  When the end of an
// output sequence is reached, playback will continue at the most-recently-output block so marked.

// g7_control_output
#define G7_FLAG_CO_START                0x00010000
#define G7_FLAG_CO_STOP                 0x00020000
#define G7_FLAG_CO_CONTINUE             0x00040000
#define G7_FLAG_CO_SPLICE               0x00080000
#define G7_FLAG_CO_TRUNCATE             0x00100000
#define G7_FLAG_CO_DISCARD              0x00200000
#define G7_FLAG_CO_START_UNSYNCHRONIZED 0x00400000

// g7_read_extended
#define G7_FLAG_READ_LATEST             0x01000000
#define G7_FLAG_READ_DISCARD            0x02000000
#define G7_FLAG_READ_24_RJ_IN_32        0x04000000

// g7_query
#define G7_FLAG_QUERY_CHANNELS          0x00000001
#define G7_FLAG_QUERY_PLAYBACK          0x00000002
#define G7_FLAG_QUERY_SETUP_C3          0x00000004

// Values for G7_FLAG_QUERY_PLAYBACK
#define G7_QUERY_PLAYBACK_UNKNOWN       0       // otherwise unused value (e.g., cannot communicate with server)
#define G7_QUERY_PLAYBACK_INITIAL       1       // codec has not yet been started.
#define G7_QUERY_PLAYBACK_PLAYING       2       // codec is playing an output stage; tag and repeat count will be valid.
#define G7_QUERY_PLAYBACK_STOPPING      3       // playback reached end of stream; waiting for server to catch up.
#define G7_QUERY_PLAYBACK_HALTED        4       // codec has been put into standby mode
#define G7_QUERY_PLAYBACK_POWERED_DOWN  5       // audio path is powered down (Otostat-C only)

// Request/response structure, sent first after opening a new connection.
// All quantities are longwords sent and received in Internet network byte order.
// On the client end, all byte ordering changes are encapsulated within issue().
struct msg_t
{
        unsigned int magic;
        unsigned int code;
        unsigned int arg1;
        unsigned int arg2;
        unsigned int arg3;
        unsigned int arg4;

        // Number of bytes of data following the header.
        unsigned int following_bytes;
};

// Magic number identifying a G7 request/response.
#define G7_MAGIC                        0x39FB443D

// Message types.
// Host to peripheral:

// Initial negotiation.
// arg1 = client protocol version (16/major, 16/minor)
// arg2 = desired number of samples per block
#define G7_MSG_INITIALIZE               0x00000001      // audio data from peripheral

// Set various parameters.
// arg1 = flag (what to set)
// arg2 = value
#define G7_MSG_CONTROL                  0x00000002

// Modify staged-for-playback stream (protocol version 1+ only)
// arg1 = flags
// arg2 = with a G7_FLAG_SO_APPEND_ flag, bitmap of which channels are non-NULL; MSB = channel 0
// arg3 = with a G7_FLAG_SO_APPEND_ flag, repeat count; otherwise not used
// arg4 = with a G7_FLAG_SO_APPEND_ flag, tag; otherwise not used
// data = with a G7_FLAG_SO_APPEND_ flag, bitcnt(arg2) channels worth of data; otherwise not used
#define G7_MSG_STAGE_OUTPUT             0x00000003

// Control playback/capture (protocol version 1+ only)
// arg1 = flags
#define G7_MSG_CONTROL_OUTPUT           0x00000004

// Retrieve data block(s)
// arg1 = channel bitmap (channel 0 = LSB)
// arg2 = block count per channel
// arg3 = flags
#define G7_MSG_READ_INPUT               0x00000005
// Response:
// arg1 = 1 if at end-of-stream, 0 otherwise
// arg2 = tag of first block returned
// arg3 = sequence number of first block returned
// data = blocks retrieved per channel

// Query various settings.
// arg1 = flags
#define G7_MSG_QUERY                    0x00000006
// Response:
// arg1..4 returned to caller.

// Issue SPI command to CP.
// data = SPI data to send.  SPI response is returned as response data.
#define G7_MSG_LINK                     0x00000007

// Peripheral to host:

// Successful response to command expecting no data
#define G7_MSG_ACK                      0x80000001
// In response to G7_MSG_INITIALIZE:
// arg1 = 16/server major version; 16/server minor version
// arg2 = 16/playback channels, 16/capture channels
// data = initialization message.

// Error response to any message
// arg1 = G_ERR_ code
// data = NUL-terminated error string.
#define G7_MSG_NAK                      0x80000002

// Successful response to command expecting data
// arg1 = for G7_MSG_READ, 1 if end-of-stream reached, 0 otherwise (end of stream indication will only occur w/no data).
// arg2 = For G7_MSG_READ, data for tag_ parameter.
// arg3 = For G7_MSG_READ, data for sequence_ parameter.
// data = always an integral number of blocks.
#define G7_MSG_DATA                     0x80000003              // audio data from peripheral

// The name of the global semaphore protecting the SPI interface to the Control Processor, and a reasonable length of time to wait for it.
#define CP_SPI_SEMAPHORE_NAME           "/cp_spi"
#define CP_SPI_SEMAPHORE_TIMEOUT_MS     100
